import numpy as np
import trimesh.parent
from pycram.tf_transformations import quaternion_from_euler

from pycram.testing import BulletWorldTestCase

from pycram.datastructures.enums import JointType, ObjectType
from pycram.datastructures.pose import PoseStamped, Point, Quaternion
from pycram.datastructures.dataclasses import Color, BoundingBox as BB
from pycram.failures import UnsupportedFileExtension
from pycram.world_concepts.world_object import Object
from pycram.object_descriptors.generic import ObjectDescription as GenericObjectDescription

import pathlib

from pycrap.ontologies import Milk, Food


class TestObject(BulletWorldTestCase):

    def test_wrong_object_description_path(self):
        with self.assertRaises(UnsupportedFileExtension):
            milk = Object("milk_not_found", Milk, "wrong_path.sk")

    def test_malformed_object_description(self):
        file_path = pathlib.Path(__file__).parent.resolve()
        malformed_file = str(file_path) + "/../resources/cached/malformed_description.urdf"
        with open(malformed_file, "w") as file:
            file.write("malformed")
        with self.assertRaises(Exception):
            Object("milk2", Milk, malformed_file)

    def test_move_base_to_origin_pose(self):
        self.milk.set_position(Point(x=1, y=2, z=3), base=False)
        self.milk.move_base_to_origin_pose()
        self.assertEqual(self.milk.get_base_position_as_list(), [1, 2, 3])

    def test_set_position_as_point(self):
        self.milk.set_position(Point(x=1, y=2, z=3))
        self.assertEqual(self.milk.get_position_as_list(), [1, 2, 3])

    def test_uni_direction_attachment(self):
        self.milk.attach(self.cereal, bidirectional=False)

        milk_position = self.milk.get_position_as_list()
        cereal_position = self.cereal.get_position_as_list()

        def move_milk_and_assert_cereal_moved():
            milk_position[0] += 1
            cereal_position[0] += 1
            self.milk.set_position(milk_position)

            new_cereal_position = self.cereal.get_position_as_list()
            self.assertEqual(new_cereal_position, cereal_position)

        def move_cereal_and_assert_milk_not_moved():
            cereal_position[0] += 1
            self.cereal.set_position(cereal_position)

            new_milk_position = self.milk.get_position_as_list()
            self.assertEqual(new_milk_position, milk_position)

        # move twice to test updated attachment at the new cereal position
        for _ in range(2):
            move_milk_and_assert_cereal_moved()
            move_cereal_and_assert_milk_not_moved()

    def test_setting_wrong_position_type(self):
        with self.assertRaises(TypeError):
            self.milk.set_position(1)

        with self.assertRaises(AssertionError):
            self.assertEqual(1, self.milk.pose.position)

    def test_set_orientation_as_list(self):
        self.milk.set_orientation([1, 0, 0, 0])
        self.assertEqual(self.milk.get_orientation_as_list(), [1, 0, 0, 0])

    def test_set_orientation_as_quaternion(self):
        self.milk.set_orientation(Quaternion(**dict(zip(["x", "y", "z", "w"], [1, 0, 0, 0]))))
        self.assertEqual(self.milk.get_orientation_as_list(), [1, 0, 0, 0])

    def test_set_orientation_as_ndarray(self):
        self.milk.set_orientation(np.array([1, 0, 0, 0]))
        self.assertEqual(self.milk.get_orientation_as_list(), [1, 0, 0, 0])

    def test_set_wrong_orientation_type(self):
        with self.assertRaises(TypeError):
            self.milk.set_orientation(1)

    def test_set_position_as_pose(self):
        self.milk.set_position(PoseStamped.from_list([1, 2, 3]))
        self.assertEqual(self.milk.get_position_as_list(), [1, 2, 3])

    def test_set_position_as_list(self):
        self.milk.set_position([1, 2, 3])
        self.assertEqual(self.milk.get_position_as_list(), [1, 2, 3])

    def test_get_joint_axis(self):
        self.assertEqual(self.robot.get_joint_axis("head_pan_joint"), Point(x=0.0, y=0.0, z=1.0))

    def test_get_joint_type(self):
        self.assertEqual(self.robot.get_joint_type("head_pan_joint"), JointType.REVOLUTE)

    def test_get_joint_lower_limit(self):
        self.assertEqual(self.robot.get_joint_lower_limit("head_pan_joint"), -3.007)

    def test_get_joint_upper_limit(self):
        self.assertEqual(self.robot.get_joint_upper_limit("head_pan_joint"), 3.007)

    def test_get_joint_damping(self):
        self.assertEqual(self.robot.get_joint_damping("head_pan_joint"), 0.5)

    def test_save_state(self):
        self.robot.attach(self.milk)
        self.robot.save_state(1)
        self.assertEqual(self.robot.saved_states[1].attachments, self.robot.attachments)
        self.assertTrue(self.milk in self.robot.saved_states[1].attachments)
        for link in self.robot.links.values():
            self.assertEqual(link.current_state, link.saved_states[1])

    def test_restore_state(self):
        self.robot.attach(self.milk)
        self.robot.save_state(1)
        self.milk.save_state(1)
        self.assertTrue(self.milk in self.robot.attachments)
        self.robot.detach(self.milk)
        self.assertTrue(self.milk not in self.robot.attachments)
        self.robot.restore_state(1)
        self.milk.restore_state(1)
        self.assertTrue(self.milk in self.robot.attachments)
        for link in self.robot.links.values():
            curr_state = link.current_state
            saved_state = link.saved_states[1]
            self.assertEqual(curr_state, saved_state)
            self.assertEqual(curr_state.constraint_ids, saved_state.constraint_ids)

    def test_get_link_by_id(self):
        self.assertEqual(self.robot.get_link_by_id(-1), self.robot.root_link)

    def test_find_joint_above_link(self):
        self.assertEqual(self.robot.find_joint_above_link("head_pan_link"), "head_pan_joint")

    def test_find_joint_above_link_exhausted(self):
        self.assertEqual(self.robot.find_joint_above_link("base_link"), None)

    def test_contact_points_simulated(self):
        self.milk.set_position([0, 0, 100])
        contact_points = self.milk.contact_points_simulated()
        self.assertFalse(contact_points)
        self.milk.set_position(self.cereal.get_position_as_list(), base=True)
        contact_points = self.milk.contact_points_simulated()
        self.assertTrue(contact_points)

    def test_set_color(self):
        self.milk.set_color(Color(1, 0, 0, 1))
        self.assertEqual(self.milk.get_color(), Color(1, 0, 0, 1))
        self.robot.set_color(Color(0, 1, 0, 1))
        for color in self.robot.get_color().values():
            self.assertEqual(color, Color(0, 1, 0, 1))

    def test_object_equal(self):
        milk2 = Object("milk2", Milk, "milk.stl")
        self.assertNotEqual(self.milk, milk2)
        self.assertEqual(self.milk, self.milk)
        self.assertNotEqual(self.milk, self.cereal)
        self.assertNotEqual(self.milk, self.world)

    def test_merge(self):
        cereal = Object("cereal2", Food, "breakfast_cereal.stl")
        milk = Object("milk2", Milk, "milk.stl")
        cereal_milk = cereal.merge(milk, new_description_file="cereal_milk")
        self.assertEqual(len(cereal_milk.links), len(cereal.links) + len(milk.links))
        self.assertEqual(len(cereal_milk.joints), len(cereal.joints) + len(milk.joints) + 1)
        cereal_milk.remove()
        # self.world.cache_manager.clear_cache()

    def test_merge_bounding_box(self):
        cereal_2 = Object("cereal2", Food, "breakfast_cereal.stl",
                          pose=self.cereal.pose)
        cereal_2.set_orientation(quaternion_from_euler(0, 0, np.pi / 2))
        cereal_bbox = self.cereal.get_axis_aligned_bounding_box(False)
        cereal_2_bbox = cereal_2.get_axis_aligned_bounding_box(False)
        plot = False
        if plot:
            BB.plot_3d_points([np.array(cereal_bbox.get_points_list()), np.array(cereal_2_bbox.get_points_list())])
        for use_random_events in [True, False]:
            merged_bbox_mesh = BB.merge_multiple_bounding_boxes_into_mesh([cereal_bbox, cereal_2_bbox],
                                                                          use_random_events=use_random_events,
                                                                          plot=plot)
            self.assertTrue(isinstance(merged_bbox_mesh, trimesh.Trimesh))
            self.assertTrue(merged_bbox_mesh.vertices.shape[0] > 0)
            self.assertTrue(merged_bbox_mesh.faces.shape[0] > 0)
            self.assertTrue(merged_bbox_mesh.is_volume if not use_random_events else True)
            self.assertTrue(merged_bbox_mesh.volume > 0 if not use_random_events else True)
        cereal_2.remove()


class GenericObjectTestCase(BulletWorldTestCase):

    def test_init_generic_object(self):
        gen_obj_desc = GenericObjectDescription("robokudo_object", [0,0,0], [0.1, 0.1, 0.1])
        obj = Object("robokudo_object", Milk, None, gen_obj_desc)
        pose = obj.get_pose()
        self.assertTrue(isinstance(pose, PoseStamped))


class OntologyIntegrationTestCase(BulletWorldTestCase):

    def test_querying(self):
        """
         Test case if spawning a second object of the same Concept in this case milk,
         Since the Object Link Description contains the same name for every object coming from the same STL-File,
         it is required to add a unique id for every new parsed link description.
         We spawn a second milk object, so there should be two instances of type milk. """
        milk2 = Object("milk2", Milk, "milk.stl")
        self.world.ontology.reason()
        r = list(filter(lambda x:  x in Milk.instances(), self.world.ontology.individuals()))
        self.assertEqual(len(r), 2)



import numpy as np
from trimesh import Trimesh
from typing_extensions import List, Tuple, Optional, Union, Dict

from pycrap.ontologies import PhysicalObject
from .datastructures.dataclasses import ContactPointsList, RayResult
from .datastructures.enums import Frame, Arms, FindBodyInRegionMethod, Grasp
from .datastructures.pose import PoseStamped, TransformStamped
from .datastructures.world import World, UseProspectionWorld
from .datastructures.world_entity import PhysicalBody
from .external_interfaces.ik import try_to_reach, try_to_reach_with_grasp
from .object_descriptors.generic import ObjectDescription as GenericObjectDescription
from .robot_description import RobotDescription, KinematicChainDescription
from .ros import logdebug, logwarn
from .utils import RayTestUtils, chunks, get_rays_from_min_max
from .world_concepts.world_object import Object, Link


def stable(obj: Object) -> bool:
    """
    Checks if an object is stable in the world. Stable meaning that it's position will not change after simulating
    physics in the World. This will be done by simulating the world for 10 seconds and compare
    the previous coordinates with the coordinates after the simulation.

    :param obj: The object which should be checked
    :return: True if the given object is stable in the world False else
    """
    prospection_obj = World.current_world.get_prospection_object_for_object(obj)
    with UseProspectionWorld():
        coords_prev = prospection_obj.get_position_as_list()
        World.current_world.set_gravity([0, 0, -9.8])

        World.current_world.simulate(2)
        coords_past = prospection_obj.get_position_as_list()

        coords_prev = list(map(lambda n: round(n, 3), coords_prev))
        coords_past = list(map(lambda n: round(n, 3), coords_past))
        return coords_past == coords_prev


def contact(
        object1: Object,
        object2: Object,
        return_links: bool = False) -> Union[bool, Tuple[bool, List[Tuple[Link, Link]]]]:
    """
    Checks if two objects are in contact or not. If the links should be returned then the output will also contain a
    list of tuples where the first element is the link name of 'object1' and the second element is the link name of
    'object2'.

    :param object1: The first object
    :param object2: The second object
    :param return_links: If the respective links on the objects that are in contact should be returned.
    :return: True if the two objects are in contact False else. If links should be returned a list of links in contact
    """

    with UseProspectionWorld():
        prospection_obj1 = World.current_world.get_prospection_object_for_object(object1)
        prospection_obj2 = World.current_world.get_prospection_object_for_object(object2)
        World.current_world.perform_collision_detection()
        con_points: ContactPointsList = World.current_world.get_contact_points_between_two_bodies(prospection_obj1,
                                                                                                  prospection_obj2)
        objects_are_in_contact = len(con_points) > 0
        if return_links:
            contact_links = [(point.body_a, point.body_b) for point in con_points]
            return objects_are_in_contact, contact_links
        else:
            return objects_are_in_contact


def prospect_robot_contact(robot: Object, ignore_collision_with: Optional[List[Object]] = None) -> bool:
    """
    Check if the robot collides with any object in the world at the given pose.

    :param robot: The robot object
    :param pose: The pose to check for collision
    :param ignore_collision_with: A list of objects to ignore collision with
    :return: True if the robot collides with any object, False otherwise
    """
    with UseProspectionWorld():
        prospection_robot = World.current_world.get_prospection_object_for_object(robot)
        pose = prospection_robot.get_pose()
        floor = prospection_robot.world.get_object_by_name("floor")
        ignore_collision_with = [] if ignore_collision_with is None else ignore_collision_with
        ignore = [o.name for o in ignore_collision_with]
        for obj in prospection_robot.world.objects:
            if obj.name in ([prospection_robot.name, floor.name] + ignore):
                continue
            in_contact, contact_links = contact(prospection_robot, obj, return_links=True)
            if in_contact and not is_held_object(prospection_robot, obj, [links[0] for links in contact_links]):
                logdebug(f"Robot is in contact with {obj.name} in prospection: {obj.world.is_prospection_world}"
                         f"at position {pose.position.to_list()} and z_angle {pose.z_angle}")
                return True
            logdebug(f"Robot is not in contact with {obj.name} in prospection: {obj.world.is_prospection_world}"
                     f"at position {pose.position.to_list()} and z_angle {pose.z_angle}")
    return False



def is_held_object(robot: Object, obj: Object, robot_contact_links: List[Link]) -> bool:
    """
    Check if the object is picked by the robot.

    :param robot: The robot object
    :param obj: The object to check if it is picked
    :param robot_contact_links: The links of the robot that are in contact with the object
    :return: True if the object is picked by the robot, False otherwise
    """
    picked_object = False
    if obj in robot.attachments:
        arm_chains = RobotDescription.current_robot_description.get_manipulator_chains()
        for chain in arm_chains:
            gripper_links = chain.end_effector.links
            if (robot.attachments[obj].parent_link.name in gripper_links
                    and all(link.name in gripper_links for link in robot_contact_links)):
                picked_object = True
                continue
    return picked_object


def get_visible_objects(
        camera_pose: PoseStamped,
        front_facing_axis: Optional[List[float]] = None,
        plot_segmentation_mask: bool = False) -> Tuple[np.ndarray, PoseStamped]:
    """
    Return a segmentation mask of the objects that are visible from the given camera pose and the front facing axis.

    :param camera_pose: The pose of the camera in world coordinate frame.
    :param front_facing_axis: The axis, of the camera frame, which faces to the front of the robot. Given as list of xyz
    :param plot_segmentation_mask: If the segmentation mask should be plotted
    :return: A segmentation mask of the objects that are visible and the pose of the point at exactly 2 meters in front of the camera in the direction of the front facing axis with respect to the world coordinate frame.
    """
    if front_facing_axis is None:
        front_facing_axis = RobotDescription.current_robot_description.get_default_camera().front_facing_axis

    camera_frame = RobotDescription.current_robot_description.get_camera_frame(World.robot.name)
    world_to_cam = camera_pose.to_transform_stamped(camera_frame)

    cam_to_point = TransformStamped.from_list(list(np.multiply(front_facing_axis, 2)), [0, 0, 0, 1], camera_frame,
                             "point")
    target_point = (world_to_cam * cam_to_point).to_pose_stamped()

    seg_mask = World.current_world.get_images_for_target(target_point, camera_pose)[2]

    if plot_segmentation_mask:
        RayTestUtils.plot_segmentation_mask(seg_mask)

    return seg_mask, target_point


def visible(
        obj: Object,
        camera_pose: PoseStamped,
        front_facing_axis: Optional[List[float]] = None,
        threshold: float = 0.8,
        plot_segmentation_mask: bool = False) -> bool:
    """
    Checks if an object is visible from a given position. This will be achieved by rendering the object
    alone and counting the visible pixel, then rendering the complete scene and compare the visible pixels with the
    absolut count of pixels.

    :param obj: The object for which the visibility should be checked
    :param camera_pose: The pose of the camera in map frame
    :param front_facing_axis: The axis, of the camera frame, which faces to the front of the robot. Given as list of xyz
    :param threshold: The minimum percentage of the object that needs to be visible for this method to return true.
    :param plot_segmentation_mask: If the segmentation mask should be plotted.
    :return: True if the object is visible from the camera_position False if not
    """
    with UseProspectionWorld():
        prospection_obj = World.current_world.get_prospection_object_for_object(obj)
        if World.robot:
            prospection_robot = World.current_world.get_prospection_object_for_object(World.robot)

        state_id = World.current_world.save_state()
        for obj in World.current_world.objects:
            if obj == prospection_obj or (World.robot and obj == prospection_robot):
                continue
            else:
                obj.set_pose(PoseStamped.from_list([100, 100, 0], [0, 0, 0, 1]), set_attachments=False)

        seg_mask, target_point = get_visible_objects(camera_pose, front_facing_axis, plot_segmentation_mask)
        max_pixel = np.array(seg_mask == prospection_obj.id).sum()

        World.current_world.restore_state(state_id)

        if max_pixel == 0:
            # Object is not visible
            return False

        seg_mask = World.current_world.get_images_for_target(target_point, camera_pose)[2]
        real_pixel = np.array(seg_mask == prospection_obj.id).sum()

        return real_pixel / max_pixel > threshold > 0


def occluding(
        obj: Object,
        camera_pose: PoseStamped,
        front_facing_axis: Optional[List[float]] = None,
        plot_segmentation_mask: bool = False) -> List[Object]:
    """
    Lists all objects which are occluding the given object. This works similar to 'visible'.
    First the object alone will be rendered and the position of the pixels of the object in the picture will be saved.
    After that the complete scene will be rendered and the previous saved pixel positions will be compared to the
    actual pixels, if in one pixel another object is visible ot will be saved as occluding.

    :param obj: The object for which occlusion should be checked
    :param camera_pose: The pose of the camera in world coordinate frame
    :param front_facing_axis: The axis, of the camera frame, which faces to the front of the robot. Given as list of xyz
    :param plot_segmentation_mask: If the segmentation mask should be plotted
    :return: A list of occluding objects
    """

    with UseProspectionWorld():
        state_id = World.current_world.save_state()
        for other_obj in World.current_world.objects:
            if other_obj.name == World.current_world.robot.name:
                continue
            elif obj.get_pose() == other_obj.get_pose():
                obj = other_obj
            else:
                other_obj.set_pose(PoseStamped.from_list([100, 100, 0], [0, 0, 0, 1]))

        seg_mask, target_point = get_visible_objects(camera_pose, front_facing_axis, plot_segmentation_mask)

        # All indices where the object that could be occluded is in the image
        # [0] at the end is to reduce by one dimension because dstack adds an unnecessary dimension
        pix = np.dstack(np.nonzero(seg_mask == obj.id))[0]

        World.current_world.restore_state(state_id)

        occluding_obj_ids = []
        seg_mask = World.current_world.get_images_for_target(target_point, camera_pose)[2]
        for c in pix:
            if not seg_mask[c[0]][c[1]] == obj.id:
                occluding_obj_ids.append(seg_mask[c[0]][c[1]])

        occ_objects = list(set(map(World.current_world.get_object_by_id, occluding_obj_ids)))
        occ_objects = list(map(World.current_world.get_object_for_prospection_object, occ_objects))

        return occ_objects


def reachable(
        pose_or_object: Union[Object, PoseStamped],
        robot: Object,
        gripper_name: str,
        threshold: float = 0.01) -> bool:
    """
    Checks if the robot can reach a given position. To determine this the inverse kinematics are
    calculated and applied. Afterward the distance between the position and the given end effector is calculated, if
    it is smaller than the threshold the reasoning query returns True, if not it returns False.

    :param pose_or_object: The position and rotation or Object for which reachability should be checked or an Object
    :param robot: The robot that should reach for the position
    :param gripper_name: The name of the end effector
    :param threshold: The threshold between the end effector and the position.
    :return: True if the end effector is closer than the threshold to the target position, False in every other case
    """

    prospection_robot = World.current_world.get_prospection_object_for_object(robot)
    with UseProspectionWorld():
        target_pose = try_to_reach(pose_or_object, prospection_robot, gripper_name)

        if not target_pose:
            return False

        gripper_pose = prospection_robot.get_link_pose(gripper_name)
        diff = target_pose.position.euclidean_distance(gripper_pose.position)

    return diff < threshold


def blocking(
        pose_or_object: Union[Object, PoseStamped],
        robot: Object,
        gripper_chain: KinematicChainDescription,
        grasp: Grasp = None) -> Union[List[Object], None]:
    """
    Checks if any objects are blocking another object when a robot tries to pick it. This works
    similar to the reachable predicate. First the inverse kinematics between the robot and the object will be
    calculated and applied. Then it will be checked if the robot is in contact with any object except the given one.
    If the given pose or Object is not reachable None will be returned

    :param pose_or_object: The object or pose for which blocking objects should be found
    :param robot: The robot Object who reaches for the object
    :param gripper_chain: The kinematic chain of the used gripper
    :param grasp: The grasp type with which the object should be grasped
    :return: A list of objects the robot is in collision with when reaching for the specified object or None if the pose or object is not reachable.
    """

    with UseProspectionWorld():
        prospection_robot = World.current_world.get_prospection_object_for_object(robot)
        if grasp:
            grasp_orientation = gripper_chain.end_effector.get_grasp(grasp, None, False)
            try_to_reach_with_grasp(pose_or_object, prospection_robot, gripper_chain.get_tool_frame(), grasp_orientation)
        else:
            try_to_reach(pose_or_object, prospection_robot, gripper_chain.get_tool_frame())

        block = [World.current_world.get_object_for_prospection_object(obj) for obj in World.current_world.objects
                 if contact(prospection_robot, obj)]
    return block


def supporting(
        object1: Object,
        object2: Object) -> bool:
    """
    Checks if one object is supporting another object. An object supports another object if they are in
    contact and the second object is above the first one. (e.g. a Bottle will be supported by a table)

    :param object1: Object that is supported
    :param object2: Object that supports the first object
    :return: True if the second object is in contact with the first one and the second is above the first else False
    """
    return contact(object1, object2) and object2.get_position().z > object1.get_position().z


def link_pose_for_joint_config(
        obj: Object,
        joint_config: Dict[str, float],
        link_name: str) -> PoseStamped:
    """
    Get the pose a link would be in if the given joint configuration would be applied to the object.
    This is done by using the respective object in the prospection world and applying the joint configuration
    to this one. After applying the joint configuration the link position is taken from there.

    :param obj: Object of which the link is a part
    :param joint_config: Dict with the goal joint configuration
    :param link_name: Name of the link for which the pose should be returned
    :return: The pose of the link after applying the joint configuration
    """
    prospection_object = World.current_world.get_prospection_object_for_object(obj)
    with UseProspectionWorld():
        for joint, pose in joint_config.items():
            prospection_object.set_joint_position(joint, pose)
        return prospection_object.get_link_pose(link_name)


def move_away_all_objects_to_create_empty_space(exclude_objects: List[str] = None):
    """
    Move all objects away from the robot to create an empty space for the robot to look at the target location.

    :param exclude_objects: The objects that should not be moved away.
    """
    exclude_objects = exclude_objects or []
    step = 10
    for i, obj in enumerate(World.current_world.objects):
        if obj.name not in exclude_objects:
            obj.set_position([100 + step * i, 100 + step * i, 0])


def generate_object_at_target(target_location: List[float], size: Tuple[float, ...] = (0.2, 0.2, 0.2),
                              name: str = "target") -> Object:
    """
    Generate a virtual object at the target location.

    :param target_location: The target location at which the object should be generated.
    :param size: The size of the object.
    :param name: The name of the object.
    """
    gen_obj_desc = GenericObjectDescription(name, [0, 0, 0], [s / 2 for s in size])
    gen_obj = Object(name, PhysicalObject, None, gen_obj_desc)
    gen_obj.set_pose(PoseStamped.from_list(target_location))
    return gen_obj


def cast_a_ray_from_camera(max_distance: float = 10):
    """
    Cast a ray from the camera to the target position.

    :param max_distance: The maximum distance the ray should be cast if no object is found.
    """
    camera_link_name = RobotDescription.current_robot_description.get_camera_link()
    camera_link = World.robot.get_link(camera_link_name)
    camera_pose = camera_link.pose
    camera_axis = RobotDescription.current_robot_description.get_default_camera().front_facing_axis
    target = np.array(camera_axis) * max_distance
    target_pose = PoseStamped.from_list(list(target), frame=camera_link.tf_frame)
    target_pose = World.robot.local_transformer.transform_pose(target_pose, Frame.Map.value)
    ray_result: RayResult = World.current_world.ray_test(camera_pose.position.to_list(),
                                                         target_pose.position.to_list())
    return ray_result


def has_gripper_grasped_body(arm: Arms, body: PhysicalBody) -> bool:
    """
    Check if the gripper has grasped the body by checking if the gripper fingers are in contact with the body.
    else it will check if the gripper links are in contact with the body.
    Note: This is different from :meth:`pycram.world_reasoning.is_held_object` as it only checks if the gripper is in
     contact with the body, not if it is attached to the gripper.

    :param arm: The arm for which the grasping should be checked.
    :param body: The body for which the grasping should be checked.
    :return: True if the gripper has grasped the body, False otherwise.
    """
    contact_links = body.get_contact_points_with_body(World.robot).get_all_bodies()
    arm_chain = RobotDescription.current_robot_description.get_arm_chain(arm)
    fingers_link_names = arm_chain.end_effector.fingers_link_names
    if fingers_link_names:
        fingers_in_contact = [link.name in fingers_link_names for link in contact_links]
        if len(fingers_in_contact) >= 2:
            return True
    else:
        logwarn(f"It is not possible to be certain of grasping if gripper fingers are not defined.")
        gripper_link_names = arm_chain.end_effector.links
        if any([link.name in gripper_link_names for link in contact_links]):
            return True
    return False


def is_body_between_fingers(body: PhysicalBody, fingers_link_names: List[str],
                            method: FindBodyInRegionMethod = FindBodyInRegionMethod.FingerToCentroid) -> bool:
    """
    Check if the body is between the fingers of the gripper.

    :param body: The body for which the check should be done.
    :param fingers_link_names: The names of the links that represent the fingers of the gripper.
    :param method: The method to use to find the body in the region.
    """
    empty_space = get_empty_space_between_fingers(fingers_link_names)
    if empty_space is None:
        return False
    intersection = get_intersection_between_body_bounding_box_and_empty_space(body, empty_space)
    if intersection is None:
        return False
    if method == FindBodyInRegionMethod.FingerToCentroid:
        # cast a ray from each finger to the centroid and check if it intersects with the object
        # this is necessary as the bounding box is usually different from the exact shape of the object
        # (e.g. concave objects)
        centroid = intersection.centroid
        for finger_name in fingers_link_names:
            finger = World.robot.links[finger_name]
            result = World.current_world.ray_test(finger.position.to_list(), centroid)
            if not (result.intersected and result.obj_id == body.id):
                return False
        return True
    else:
        return is_body_in_region(body, intersection, method=method)


def get_empty_space_between_fingers(fingers_link_names: List[str]) -> Optional[Trimesh]:
    """
    Check if there is empty space between the fingers of the gripper.

    :param fingers_link_names: The names of the links that represent the fingers of the gripper.
    :return: The empty space between the fingers.
    """
    fingers_links = [World.robot.links[link_name] for link_name in fingers_link_names]
    fingers_chs: List[Trimesh] = [link.get_convex_hull() for link in fingers_links]
    fingers_only_ch = fingers_chs[0].union(fingers_chs[1:])
    fingers_plus_empty_space = fingers_only_ch.convex_hull
    empty_space_between_fingers = fingers_plus_empty_space.difference(fingers_only_ch)
    # 10 cm^3 (if we assume a finger area of at least 2 cm^2, this means we allow at least 5 cm distance between
    # fingers).
    if empty_space_between_fingers.volume * 10 ** 6 < 10:
        return None
    return empty_space_between_fingers


def get_intersection_between_body_bounding_box_and_empty_space(body: PhysicalBody, empty_space: Trimesh) \
        -> Optional[Trimesh]:
    """
    Get the intersection between the body and the empty space between the fingers.

    :param body: The body for which the intersection should be calculated.
    :param empty_space: The empty space between the fingers.
    :return: The intersection between the body and the empty space between the fingers.
    """
    body_bb = body.get_axis_aligned_bounding_box()
    try:
        intersection = empty_space.intersection(body_bb.as_mesh)
    except ValueError:
        # This will be raised when the empty_space has no volume.
        return None
    if intersection.volume * 10 ** 6 < 10:
        return None
    return intersection


def is_body_in_region(body: PhysicalBody, region: Trimesh, step_size_in_meters: float = 0.01,
                      method: FindBodyInRegionMethod = FindBodyInRegionMethod.Centroid) -> bool:
    """
    Check if the body is in the given region.

    :param body: The body for which the check should be done.
    :param region: The region to check if the body is in.
    :param step_size_in_meters: The step size in meters between the rays.
    :param method: The method to use to find the body in the region.
    """
    min_bound, max_bound = region.bounds
    reduction = 0.002
    min_bound = min_bound + np.array([reduction]*3)
    max_bound = max_bound - np.array([reduction]*3)
    if method == FindBodyInRegionMethod.Centroid:
        centroid = region.centroid
        ray1 = World.current_world.ray_test(min_bound, centroid)
        ray2 = World.current_world.ray_test(max_bound, centroid)
        return any([ray.intersected and ray.obj_id == body.id for ray in [ray1, ray2]])
    elif method == FindBodyInRegionMethod.MultiRay:
        # cast multiple rays with small steps starting from min_bound to max_bound
        rays = get_rays_from_min_max(min_bound, max_bound, step_size_in_meters)
        max_batch_size = World.current_world.conf.max_batch_size_for_rays
        max_batch_size = max_batch_size if max_batch_size is not None else rays.shape[0]
        for n in chunks(rays, max_batch_size):
            ray_results = World.current_world.ray_test_batch(n[:, :, 0].reshape((-1, 3)), n[:, :, 1].reshape((-1, 3)))
            obj_ids = [result.obj_id for result in ray_results if result.intersected]
            if body.id in obj_ids:
                return True
        return False



from ..robot_description import RobotDescription, KinematicChainDescription, EndEffectorDescription, \
    RobotDescriptionManager, CameraDescription
from ..datastructures.enums import Arms, Grasp, GripperState, GripperType, TorsoState, StaticJointState
from ..ros import get_ros_package_path
from ..datastructures.dataclasses import VirtualMobileBaseJoints

filename = get_ros_package_path('pycram') + '/resources/robots/' + "Armar6" + '.urdf'

armar_description = RobotDescription("Armar6", "world", "torso", "torso_joint",
                                      filename, virtual_mobile_base_joints=VirtualMobileBaseJoints())

################################## Left Arm ##################################
left_arm = KinematicChainDescription("left", "world", "arm_t8_r0",
                                     armar_description.urdf_object, arm_type=Arms.LEFT)

left_arm.add_static_joint_states(StaticJointState.Park, {"torso_joint": -0.15,
                                          "arm_t12_joint_r0": 0,
                                          "arm_t23_joint_r0": 0,
                                          "arm_t34_joint_r0": 1.5,
                                          "arm_t45_joint_r0": 0.5,
                                          "arm_t56_joint_r0": 2.0,
                                          "arm_t67_joint_r0": 1.5,
                                          "arm_t78_joint_r0": 0,
                                          "arm_t8_joint_r0": 0})

armar_description.add_kinematic_chain_description(left_arm)

################################## Left Gripper ##################################
left_gripper = EndEffectorDescription("left_gripper", "arm_t8_r0", "left_tool_frame",
                                      armar_description.urdf_object)
left_gripper.add_static_joint_states(GripperState.OPEN, {"Thumb L 1 Joint": 0.0,
                                                         "Thumb L 2 Joint": 0.0,
                                                         "Index L 1 Joint": 0.0,
                                                         "Index L 2 Joint": 0.0,
                                                         "Index L 3 Joint": 0.0,
                                                         "Middle L 1 Joint": 0.0,
                                                         "Middle L 2 Joint": 0.0,
                                                         "Middle L 3 Joint": 0.0,
                                                         "Ring L 1 Joint": 0.0,
                                                         "Ring L 2 Joint": 0.0,
                                                         "Ring L 3 Joint": 0.0,
                                                         "Pinky L 1 Joint": 0.0,
                                                         "Pinky L 2 Joint": 0.0,
                                                         "Pinky L 3 Joint": 0.0})
left_gripper.add_static_joint_states(GripperState.CLOSE, {"Thumb L 1 Joint": 1.57,
                                                          "Thumb L 2 Joint": 1.57,
                                                          "Index L 1 Joint": 1.57,
                                                          "Index L 2 Joint": 1.57,
                                                          "Index L 3 Joint": 1.57,
                                                          "Middle L 1 Joint": 1.57,
                                                          "Middle L 2 Joint": 1.57,
                                                          "Middle L 3 Joint": 1.57,
                                                          "Ring L 1 Joint": 1.57,
                                                          "Ring L 2 Joint": 1.57,
                                                          "Ring L 3 Joint": 1.57,
                                                          "Pinky L 1 Joint": 1.57,
                                                          "Pinky L 2 Joint": 1.57,
                                                          "Pinky L 3 Joint": 1.57})

left_gripper.end_effector_type = GripperType.FINGER
# left_gripper.opening_distance = 0.548
left_arm.end_effector = left_gripper

################################## Right Arm ##################################
right_arm = KinematicChainDescription("right", "world", "arm_t8_r1",
                                      armar_description.urdf_object, arm_type=Arms.RIGHT)

right_arm.add_static_joint_states(StaticJointState.Park, {"torso_joint": -0.15,
                                           "arm_t12_joint_r1": 0,
                                           "arm_t23_joint_r1": 0,
                                           "arm_t34_joint_r1": 1.5,
                                           "arm_t45_joint_r1": 2.64,
                                           "arm_t56_joint_r1": 2.0,
                                           "arm_t67_joint_r1": 1.6415,
                                           "arm_t78_joint_r1": 0,
                                           "arm_t8_joint_r1": 0})

armar_description.add_kinematic_chain_description(right_arm)

################################## Right Gripper ##################################
right_gripper = EndEffectorDescription("right_gripper", "arm_t8_r1", "right_tool_frame",
                                       armar_description.urdf_object)
right_gripper.add_static_joint_states(GripperState.OPEN, {"Thumb R 1 Joint": 0.0,
                                                          "Thumb R 2 Joint": 0.0,
                                                          "Index R 1 Joint": 0.0,
                                                          "Index R 2 Joint": 0.0,
                                                          "Index R 3 Joint": 0.0,
                                                          "Middle R 1 Joint": 0.0,
                                                          "Middle R 2 Joint": 0.0,
                                                          "Middle R 3 Joint": 0.0,
                                                          "Ring R 1 Joint": 0.0,
                                                          "Ring R 2 Joint": 0.0,
                                                          "Ring R 3 Joint": 0.0,
                                                          "Pinky R 1 Joint": 0.0,
                                                          "Pinky R 2 Joint": 0.0,
                                                          "Pinky R 3 Joint": 0.0})
right_gripper.add_static_joint_states(GripperState.CLOSE, {"Thumb R 1 Joint": 1.57,
                                                           "Thumb R 2 Joint": 1.57,
                                                           "Index R 1 Joint": 1.57,
                                                           "Index R 2 Joint": 1.57,
                                                           "Index R 3 Joint": 1.57,
                                                           "Middle R 1 Joint": 1.57,
                                                           "Middle R 2 Joint": 1.57,
                                                           "Middle R 3 Joint": 1.57,
                                                           "Ring R 1 Joint": 1.57,
                                                           "Ring R 2 Joint": 1.57,
                                                           "Ring R 3 Joint": 1.57,
                                                           "Pinky R 1 Joint": 1.57,
                                                           "Pinky R 2 Joint": 1.57,
                                                           "Pinky R 3 Joint": 1.57})

right_gripper.end_effector_type = GripperType.FINGER
# right_gripper.opening_distance = 0.548
right_arm.end_effector = right_gripper

################################## Torso ##################################
torso = KinematicChainDescription("torso", "platform", "torso",
                                  armar_description.urdf_object)

torso.add_static_joint_states(TorsoState.HIGH, {"torso_joint": 0.0})
torso.add_static_joint_states(TorsoState.MID, {'torso_joint': -0.185})
torso.add_static_joint_states(TorsoState.LOW, {"torso_joint": -0.365})

armar_description.add_kinematic_chain_description(torso)

################################## Camera ##################################
camera = CameraDescription("Roboception", "Roboception", 1.371500015258789,
                           1.7365000247955322, 0.99483, 0.75049,
                           [0, 0, 1])
armar_description.add_camera_description(camera)

################################## Neck ##################################
armar_description.add_kinematic_chain("neck", "lower_neck", "upper_neck")
armar_description.set_neck(yaw_joint="neck_1_yaw", pitch_joint="neck_2_pitch")


################################# Grasps ##################################
orientation = [0.707, 0.707, 0.707, 0.707]
right_gripper.update_all_grasp_orientations(orientation)
left_gripper.update_all_grasp_orientations(orientation)


################################# Additionals ##################################
# armar_description.set_costmap_offset(0)
# armar_description.set_max_reach("torso", "left_tool_frame")
# armar_description.set_palm_axis([0, 0, 1])

# Add to RobotDescriptionManager
rdm = RobotDescriptionManager()
rdm.register_description(armar_description)



from ..ros import  get_ros_package_path

from ..robot_description import RobotDescription, KinematicChainDescription, EndEffectorDescription, \
    RobotDescriptionManager, CameraDescription
from ..datastructures.enums import GripperState, Grasp, Arms, TorsoState, GripperType, StaticJointState
from ..units import meter


filename = get_ros_package_path('pycram') + '/resources/robots/' + "hsrb" + '.urdf'

hsrb_description = RobotDescription("hsrb", "base_link", "arm_lift_link", "arm_lift_joint",
                                    filename)

################################## Left Arm ##################################
left_arm = KinematicChainDescription("left_arm", "arm_lift_link", "hand_palm_link",
                                     hsrb_description.urdf_object, arm_type=Arms.LEFT)

left_arm.add_static_joint_states(StaticJointState.Park, {'arm_flex_joint': 0.0,
                                          'arm_roll_joint': 1.5,
                                          'wrist_flex_joint': -1.85,
                                          'wrist_roll_joint': 0.0})
hsrb_description.add_kinematic_chain_description(left_arm)

################################## Left Gripper ##################################
left_gripper = EndEffectorDescription("left_gripper", "wrist_roll_link", "hand_gripper_tool_frame",
                                      hsrb_description.urdf_object)

left_gripper.add_static_joint_states(GripperState.OPEN, {'hand_l_proximal_joint': 0.3,
                                                         'hand_r_proximal_joint': 0.3,
                                                         'hand_motor_joint': 0.3})
left_gripper.add_static_joint_states(GripperState.CLOSE, {'hand_l_proximal_joint': 0.0,
                                                          'hand_r_proximal_joint': 0.0,
                                                          'hand_motor_joint': 0.0})
left_gripper.end_effector_type = GripperType.PARALLEL
left_gripper.opening_distance = 0.13 * meter
left_arm.end_effector = left_gripper

################################## Torso ##################################
torso = KinematicChainDescription("torso", "base_link", "torso_lift_link",
                                  hsrb_description.urdf_object)

torso.add_static_joint_states(TorsoState.HIGH, {"torso_lift_joint": 0.34})

torso.add_static_joint_states(TorsoState.MID, {"torso_lift_joint": 0.17})

torso.add_static_joint_states(TorsoState.LOW, {"torso_lift_joint": 0})

hsrb_description.add_kinematic_chain_description(torso)

################################## Camera ##################################

head_center_camera = CameraDescription("head_center_camera_frame", "head_center_camera_frame", 0.99483, 0.75049)
head_r_camera = CameraDescription("head_r_stereo_camera_link", "head_r_stereo_camera_link", 0.99483, 0.75049)
head_l_camera = CameraDescription("head_l_stereo_camera_link", "head_l_stereo_camera_link", 0.99483, 0.75049)
head_rgbd_camera = CameraDescription("head_rgbd_sensor_link", "head_rgbd_sensor_link", 0.99483, 0.75049)
hand_camera = CameraDescription("hand_camera_frame", "hand_camera_frame", 0.99483, 0.75049)

hsrb_description.add_camera_description(head_center_camera)
hsrb_description.add_camera_description(head_r_camera)
hsrb_description.add_camera_description(head_l_camera)
hsrb_description.add_camera_description(head_rgbd_camera)
hsrb_description.add_camera_description(hand_camera)

################################## Neck ##################################
neck = KinematicChainDescription("neck", "head_pan_link", "head_tilt_link",
                                 hsrb_description.urdf_object)
hsrb_description.set_neck(yaw_joint="head_pan_joint", pitch_joint="head_tilt_joint")

################################# Grasps ##################################
left_gripper.update_all_grasp_orientations([-1, 0, -1, 0])

hsrb_description.add_kinematic_chain_description(neck)

# Add to RobotDescriptionManager
rdm = RobotDescriptionManager()
rdm.register_description(hsrb_description)



from __future__ import annotations

from abc import abstractmethod, ABC

from owlready2 import Imp, Ontology
from typing_extensions import Type, Dict

from .object_properties import ontology as CRAXOntology, is_part_of, contains_object, CRAX_ONTOLOGY_NAME


class CRAXRule(ABC):
    """
    A class that represents a rule in PyCRAP.
    """
    all_rules: Dict[Ontology, Dict[Type[CRAXRule], CRAXRule]] = {}
    """
    A dictionary that maps ontology rule classes to rule instances, so that the same rule is not created multiple times.
    """

    def __init__(self, ontology: Ontology = CRAXOntology,
                 ontology_name: str = CRAX_ONTOLOGY_NAME):
        """
        Creates a new CRAXRule object and sets the rule in the ontology.

        :param ontology: The ontology to add the rule to.
        :param ontology_name: The name of the ontology.
        """
        self.ontology = ontology
        self.ontology_name = ontology_name
        if self.rule_in_all_rules:
            self.rule = self._get_old_rule()
        else:
            self.rule = self._add_rule()

    @property
    def rule_in_all_rules(self):
        return self.ontology in CRAXRule.all_rules and type(self) in CRAXRule.all_rules[self.ontology]

    def _get_old_rule(self):
        """
        Gets the old rule from the all_rules dictionary.
        """
        return CRAXRule.all_rules[self.ontology][type(self)].rule

    def _add_rule(self):
        """
        Adds the rule to the ontology.
        """
        rule = Imp(namespace=self.ontology)
        rule_body = self.rule_body.replace(f"{self.ontology_name}.", "")
        rule.set_as_rule(rule_body)
        return rule

    @property
    def name(self):
        """
        :return: The name of the rule
        """
        return self.__class__.__name__

    @property
    @abstractmethod
    def rule_body(self):
        """
        :return: The body of the rule as a string in SWRL syntax.
        """
        pass


class HierarchicalContainment(CRAXRule):
    """
    A rule that asserts that parent objects contain the objects contained by their parts.
    """

    @property
    def rule_body(self):
        return f"""
        {is_part_of}(?part, ?parent), {contains_object}(?part, ?object)
        -> {contains_object}(?parent, ?object)
        """



from enum import Enum, auto
from functools import cached_property

import numpy as np
from .tf_transformations import quaternion_from_euler
from random_events.interval import closed_open
from typing_extensions import Optional, Type

from .datastructures.world import World
from .costmaps import Costmap, OccupancyCostmap, VisibilityCostmap
import matplotlib.colorbar
from .datastructures.pose import PoseStamped
from .ros import create_publisher, Duration, loginfo
from .units import meter

from pint import Quantity
from probabilistic_model.probabilistic_circuit.nx.helper import uniform_measure_of_event
from probabilistic_model.probabilistic_circuit.nx.probabilistic_circuit import ProbabilisticCircuit
from random_events.product_algebra import Event, SimpleEvent
from random_events.variable import Continuous

try:
    from std_msgs.msg import ColorRGBA
    from visualization_msgs.msg import Marker, MarkerArray
except ImportError:
    loginfo("Could not import visualization_msgs.msg.Marker and std_msgs.msg.ColorRGBA. "
            "This is probably because you are not running ROS.")



class Filter(Enum):
    OCCUPANCY = auto()
    VISIBILITY = auto()


class ProbabilisticCostmap:
    """
    A Costmap that uses probability distributions for representation.
    """

    x: Continuous = Continuous("x")
    """
    The variable for the x-axis (height) in meters.
    """

    y: Continuous = Continuous("y")
    """
    The variable for the y-axis (width) in meters.
    """

    costmap: Costmap
    """
    The legacy costmap.
    """

    origin: PoseStamped
    """
    The origin of the costmap.
    """

    size: Quantity
    """
    The side length of the costmap. The costmap is a square.
    """

    distribution: Optional[ProbabilisticCircuit] = None
    """
    The distribution associated with the costmap.
    """

    def __init__(self, origin: PoseStamped,
                 size: Quantity = 2 * meter,
                 max_cells = 10000,
                 costmap_type: Type[Costmap] = OccupancyCostmap,
                 world: Optional[World] = None):

        self.world = world if world else World.current_world
        self.origin = origin
        self.size = size

        # calculate the number of cells per axis
        number_of_cells = int(np.sqrt(max_cells))
        resolution = self.size.to(meter) / number_of_cells

        if costmap_type == OccupancyCostmap:
            robot_bounding_box = self.world.robot.get_axis_aligned_bounding_box()
            distance_to_obstacle = max(robot_bounding_box.width, robot_bounding_box.depth) / 2
            self.costmap = OccupancyCostmap(
                origin=self.origin,
                distance_to_obstacle=distance_to_obstacle,
                size=number_of_cells,
                resolution=resolution.magnitude,
                from_ros=False,
                world = self.world)
        elif costmap_type == VisibilityCostmap:
            camera = list(self.world.robot_description.cameras.values())[0]
            self.costmap = VisibilityCostmap(
                min_height=camera.minimal_height, max_height=camera.maximal_height, size=number_of_cells,
                resolution=resolution.magnitude, origin=self.origin, world=self.world)
        else:
            raise NotImplementedError(f"Unknown costmap type {costmap_type}")
        self.create_distribution()

    @cached_property
    def publisher(self):
        return create_publisher("/pycram/viz_marker", MarkerArray, queue_size=10)

    def create_event_from_map(self) -> Event:
        """
        :return: The event that is encoded by the costmaps map.
        """
        area = Event()
        for rectangle in self.costmap.partitioning_rectangles():
            rectangle.translate(self.origin.position.x, self.origin.position.y)
            area.simple_sets.add(SimpleEvent({self.x: closed_open(rectangle.x_lower, rectangle.x_upper),
                                                    self.y: closed_open(rectangle.y_lower, rectangle.y_upper)}))
        return area

    def create_distribution(self):
        """
        Create a probabilistic circuit from the costmap.
        """
        # self.distribution = fully_factorized([self.x, self.y], {self.x: self.origin.position.x,
        #                                                     self.y: self.origin.position.y},
        #                                      {self.x: 1, self.y: 1})
        # self.distribution, _ = self.distribution.conditional(self.create_event_from_map())
        self.distribution = uniform_measure_of_event(self.create_event_from_map())

    def sample_to_pose(self, sample: np.ndarray) -> PoseStamped:
        """
        Convert a sample from the costmap to a pose.

        :param sample: The sample to convert
        :return: The pose corresponding to the sample
        """
        x = sample[0]
        y = sample[1]
        position = [x, y, self.origin.position.z]
        angle = np.arctan2(position[1] - self.origin.position.y, position[0] - self.origin.position.x) + np.pi
        orientation = list(quaternion_from_euler(0, 0, angle, axes="sxyz"))
        return PoseStamped.from_list(position, orientation, self.origin.frame_id)

    def visualize(self):
        """
        Visualize the costmap for rviz.
        """
        samples = self.distribution.sample(1000)
        likelihoods = self.distribution.likelihood(samples)
        likelihoods /= max(likelihoods)

        colorscale = matplotlib.cm.get_cmap("inferno")
        marker = Marker()
        marker.type = Marker.POINTS
        marker.id = 0
        marker.action = Marker.ADD
        marker.header.frame_id = self.origin.frame_id
        marker.pose = PoseStamped().pose
        marker.lifetime = Duration(60)
        marker.scale.x = 0.05
        marker.scale.y = 0.05


        for index, (sample, likelihood) in enumerate(zip(samples, likelihoods)):
            position = self.sample_to_pose(sample).pose.position
            position.z = 0.1
            marker.points.append(position)
            marker.colors.append(ColorRGBA(**dict(zip(["r", "g", "b","a"], [*colorscale(likelihood)[:3], 1.0]))))

        marker_array = MarkerArray()
        marker_array.markers.append(marker)
        self.publisher.publish(marker_array)



from logging import getLogger, Logger
from functools import lru_cache
import logging

format = '%(filename)s %(lineno)s %(funcName)s %(message)s'
logging.basicConfig(format=format)
logger = getLogger(__name__)

def logwarn(msg: str):
    logger.warning(msg)

def logerr(msg: str):
    logger.error(msg)

def loginfo(msg: str):
    logger.info(msg)

def logdebug(msg: str):
    logger.debug(msg)

@lru_cache(None)
def logwarn_once(msg: str):
    logger.warning(msg)

@lru_cache(None)
def logerr_once(msg: str):
    logger.error(msg)

@lru_cache(None)
def loginfo_once(msg: str):
    logger.info(msg)

@lru_cache(None)
def logdebug_once(msg: str):
    logger.debug(msg)

def set_logger_level(level: int):
    """
    Set the logging level for the logger.
    :param level: The logging level to set.
    """
    logger.setLevel(level)


from unittest import TestCase, skipIf
try:
    from pycram.object_descriptors.mjcf import ObjectDescription as MJCFObjDesc
    from dm_control import mjcf
except ImportError:
    MJCFObjDesc = None


@skipIf(MJCFObjDesc is None, "Multiverse not found.")
class TestMjcf(TestCase):
    model: MJCFObjDesc

    @classmethod
    def setUpClass(cls):
        # Example usage
        model = mjcf.RootElement("test")

        model.default.dclass = 'default'

        # Define a simple model with bodies and joints
        body1 = model.worldbody.add('body', name='body1')
        body2 = body1.add('body', name='body2')
        joint1 = body2.add('joint', name='joint1', type='hinge')

        body3 = body2.add('body', name='body3')
        joint2 = body3.add('joint', name='joint2', type='slide')

        cls.model = MJCFObjDesc()
        cls.model.update_description_from_string(model.to_xml_string())

    def test_child_map(self):
        self.assertEqual(self.model.child_map, {'body1': [('joint1', 'body2')], 'body2': [('joint2', 'body3')]})

    def test_parent_map(self):
        self.assertEqual(self.model.parent_map, {'body2': ('joint1', 'body1'), 'body3': ('joint2', 'body2')})

    def test_get_chain(self):
        self.assertEqual(self.model.get_chain('body1', 'body3'),
                         ['body1', 'joint1', 'body2', 'joint2', 'body3'])




from enum import EnumMeta
from owlready2 import get_ontology, Thing, DataProperty, types
from random_events.utils import recursive_subclasses
from random_events.utils import get_full_class_name
from typing_extensions import Dict, List, Type, Optional, Any, get_origin, Union, get_args
from dataclasses import dataclass
from pathlib import Path
import inspect
import ast
import os
import re

from ..designator import ActionDescription


@dataclass
class ParameterDigest:
    """
    Encapsulation of meta information about a parameter.
    """
    clazz: Any
    """
    Class of the parameter.
    """
    name: str
    """
    Name of the parameter.
    """
    docstring: str
    """
    Docstring of the parameter itself (individual to each performable).
    """
    default_value: Optional[Any]
    """
    Holds the default value of the parameter if set.
    """

    @property
    def docstring_of_clazz(self) -> str: return self.clazz.__doc__

    @property
    def is_enum(self) -> bool: return self.clazz.__class__ == EnumMeta

    @property
    def is_optional(self) -> bool: return (type(None) in get_args(self.clazz)) if get_origin(
        self.clazz) is Union else False

    def get_default_value(self) -> Optional[List[str]]:
        """
        :return: A list containing the string representation of the default value or
            `None` if no default value exists.
        """
        return None if self.default_value == inspect.Parameter.empty else [str(self.default_value)]


class ActionAbstractDigest:
    """
    Wrap all information about an action abstract class that are necessary for the created ontology.
    """

    def __init__(self, clazz: Type[ActionDescription]):
        self.clazz: Type[ActionDescription] = clazz
        self.full_name: str = get_full_class_name(clazz)
        self.classname: str = clazz.__name__
        self.docstring: str = clazz.__doc__ or ""
        self.parameters: Optional[List[ParameterDigest]] = self.extract_dataclass_parameter_information()

    def extract_dataclass_parameter_information(self) -> List[ParameterDigest]:
        """
        Extract information about dataclass parameters from a dataclass.

        :return: List of parameter information.
        """

        with open(inspect.getfile(self.clazz), 'r') as file:
            file_content = file.read()
        tree = ast.parse(file_content)
        class_param_comment: Dict = {}
        for node in ast.walk(tree):
            if isinstance(node, ast.ClassDef) and node.name == self.clazz.__name__:
                last_assign: List = []
                for item in node.body:
                    if isinstance(item, ast.AnnAssign):
                        last_assign = [item.target.id]
                        class_param_comment[item.target.id] = ""
                    elif isinstance(item, ast.Assign):
                        last_assign = list(map(lambda tar: tar.id, list(item.targets)))
                        class_param_comment = {**class_param_comment, **{var: "" for var in last_assign}}
                    elif last_assign and isinstance(item, ast.Expr):
                        class_param_comment = {**class_param_comment, **{var: item.value.s for var in last_assign}}
                        last_assign = []

        parameters_inspection = inspect.signature(self.clazz).parameters
        parameter_digests = []
        for param in list(parameters_inspection.keys()):
            param_clazz = self.clazz.get_type_hints()[param]
            parameter_digests.append(
                ParameterDigest(
                    clazz=param_clazz,
                    name=param,
                    docstring=class_param_comment[param] if param in class_param_comment.keys() else "",
                    default_value=parameters_inspection[param].default,
                ))
        return parameter_digests


def create_ontology_from_performables(
        output_path: Path = "./performables.owl",
        abstract_actions_to_parse: Union[List[Type[ActionDescription]], Type[ActionDescription]] = None) -> None:
    """
    Create an ontology from the performables.

    :param output_path: Path of the output ontology file.
    :param abstract_actions_to_parse: ActionAbstract classes to parse.
    If not set, all subclasses of ActionAbstract will be parsed.
    """

    def unwrap_classname(parameter: ParameterDigest) -> str:
        """
        Unwrap the class name of the type of parameter.
        When it is an optional type, the class name of the optional type is returned.

        :param parameter: ParameterDigest to return the unwrapped classname from.
        :return: Unwrapped classname of the parameter.
        """

        def extract_content_between_quotes(text: str) -> str:
            """
            :param text: String with quotation marks.
            :return: Text without quotation marks.
            """
            if match := re.search(r"'(.*?)'", text):
                return match.group(1)
            else:
                return text

        def get_optional_type(t) -> Optional[str]:
            """
            :param t: Type hint to check.
            :return: Full class name of the optional type or None if not optional.
            """
            optional_types = [arg for arg in get_args(t) if arg is not type(None)]
            if len(optional_types) > 1:
                print(f"Optional type has more than one type: {optional_types} (Type: {t})")
            return get_full_class_name(optional_types[0]) if len(optional_types) >= 1 else None

        clazz = parameter.clazz
        if parameter.is_optional:
            clazz = get_optional_type(parameter.clazz)
        return extract_content_between_quotes(str(clazz)).replace(" ", "")

    # If parameter is not set, all subclasses of ActionAbstract will be parsed.

    if abstract_actions_to_parse:
        try:
            iter(abstract_actions_to_parse)
        except TypeError:
            abstract_actions_to_parse = [abstract_actions_to_parse]
    else:
        abstract_actions_to_parse = recursive_subclasses(ActionDescription)

    classes = [ActionAbstractDigest(clazz) for clazz in abstract_actions_to_parse]

    # Definition of created ontology
    output_ontology = get_ontology("performables")
    with output_ontology:
        class Performable(Thing): pass

        class Parameter(Thing): pass

        class Enum(Thing): pass

        class has_parameter(Performable >> Parameter): pass

        class has_default_value(DataProperty): pass

        class has_possible_value(Parameter >> Enum): pass

        class is_optional(Parameter >> bool): pass

        class has_description(DataProperty):
            range = [str]

    def create_parameter_onto_class(param: ParameterDigest) -> Parameter:
        """
        Create the ontology classes for the parameter classes.

        :param param: ParameterDigest to create the ontology classes for.
        """

        def create_enum_onto_class_and_instances() -> None:
            """
            Create ontology classes for enums and instances for possible values of enums.
            """
            enum_value_class = types.new_class(classname + "_Value", (Enum,))
            parameter_clazz.has_possible_value = [enum_value_class]
            # Create the possible values of the enum as instances of the Enum class
            for enum_member in param.clazz.__members__:
                enum_value_class(enum_member)

        parameter_clazz = types.new_class(classname, (Parameter,))
        parameter_clazz.has_description = param.docstring_of_clazz
        if param.is_enum:
            create_enum_onto_class_and_instances()
        return parameter_clazz

    def create_performable_onto_class() -> None:
        """
        Create an instance of the ontology class Performable for each ActionAbstract and all the relations to its
        ontological parameter classes.
        """

        def create_param_onto_instances() -> List[Parameter]:
            """
            Create ontology instances for the parameters of the Performable.
            """
            params = []
            for param_digest in clazz_digest.parameters:
                param_instance = all_param_classes_to_ontological_class[unwrap_classname(param_digest)](
                    param_digest.name)
                param_instance.has_description = [param_digest.docstring]
                param_instance.is_optional = [True] if param_digest.is_optional else [False]
                if param_digest.get_default_value():
                    param_instance.has_default_value = param_digest.get_default_value()
                params.append(param_instance)
            return params

        performable = Performable(clazz_digest.classname)
        performable.has_description = [clazz_digest.docstring]
        performable.has_parameter = create_param_onto_instances()

    # Creation of ontological classes/instances for the parsed Python classes.
    all_param_classes_to_ontological_class = {}
    for clazz_digest in classes:
        for param in clazz_digest.parameters:
            if (classname := unwrap_classname(param)) not in all_param_classes_to_ontological_class.keys():
                all_param_classes_to_ontological_class[classname] = create_parameter_onto_class(param)

    # Creating the ontology instances based on the created ActionAbstractDigests.
    for clazz_digest in classes:
        create_performable_onto_class()

    (dirname, filename) = os.path.split(output_path)
    os.makedirs(dirname, exist_ok=True)
    output_ontology.save(file=output_path, format="rdfxml")



from __future__ import annotations

import copy
import datetime
import math
from dataclasses import dataclass, field, fields

import numpy as np
from scipy.spatial.transform import Rotation as R
from typing_extensions import Self, Tuple, Optional, List, TYPE_CHECKING, Any

from .enums import AxisIdentifier, Arms, Grasp
from .grasp import GraspDescription, PreferredGraspAlignment
from ..has_parameters import has_parameters, HasParameters
from ..ros import Time as ROSTime
from ..tf_transformations import quaternion_multiply, translation_matrix, quaternion_matrix, inverse_matrix, \
    translation_from_matrix, quaternion_from_matrix

if TYPE_CHECKING:
    from ..world_concepts.world_object import Object


@has_parameters
@dataclass
class Vector3(HasParameters):
    """
    A 3D vector with x, y and z coordinates.
    """

    x: float = 0
    y: float = 0
    z: float = 0

    def euclidean_distance(self, other: Self) -> float:
        """
        The euclidian distance between this vector and another vector.

        :param other: The other vector to calculate the distance to.
        :return: The euclidian distance
        """
        return ((self.x - other.x) ** 2 + (self.y - other.y) ** 2 + (self.z - other.z) ** 2) ** 0.5

    def ros_message(self):
        """
        Convert the vector to a ROS message of type Vector3.

        :return: The ROS message.
        """
        from geometry_msgs.msg import Vector3 as ROSVector3
        return ROSVector3(x=float(self.x), y=float(self.y), z=float(self.z))

    def to_list(self) -> List[float]:
        """
        Convert the vector to a list.

        :return: A list containing the x, y and z coordinates.
        """
        return [self.x, self.y, self.z]

    def round(self, decimals: int = 4):
        """
        Rounds the coordinates of the vector to the specified number of decimal places.

        :param decimals: Number of decimal places to round to.
        """
        self.x = round(self.x, decimals)
        self.y = round(self.y, decimals)
        self.z = round(self.z, decimals)

    def almost_equal(self, other: Self, tolerance: float = 1e-6) -> bool:
        """
        Check if two vectors are almost equal within a given tolerance.

        :param other: The other vector to compare to.
        :param tolerance: The tolerance for the comparison as number of decimal places.
        :return: True if the vectors are almost equal, False otherwise.
        """
        return bool(np.isclose(np.array(self.to_list()), np.array(other.to_list()), atol=tolerance).all())

    def vector_to_position(self, other: Self) -> Vector3:
        """
        Calculates a vector from this vector to another vector.

        :param other: The vector to calculate the vector to.
        :return: A new vector between this vector and the other vector.
        """
        return other - self

    def to_numpy(self) -> np.ndarray:
        """
        Convert the vector to a numpy array.

        :return: A numpy array containing the x, y and z coordinates.
        """
        return np.array(self.to_list())

    def __add__(self, other: Self) -> Vector3:
        """
        Adds two vectors together.

        :param other: The other vector to add.
        :return: A new vector that is the sum of this vector and the other vector.
        """
        return Vector3(self.x + other.x, self.y + other.y, self.z + other.z)

    def __sub__(self, other: Self) -> Vector3:
        """
        Subtracts two vectors.

        :param other: The other vector to subtract.
        :return: A new vector that is the difference of this vector and the other vector.
        """
        return Vector3(self.x - other.x, self.y - other.y, self.z - other.z)

    def __mul__(self, other: float) -> Vector3:
        """
        Multiplies the vector by a scalar.

        :param other: The scalar to multiply by.
        :return: A new vector that is the product of this vector and the scalar.
        """
        return Vector3(self.x * other, self.y * other, self.z * other)

    def __rmul__(self, other: float) -> Vector3:
        """
        Multiplies the vector by a scalar (right multiplication).

        :param other: The scalar to multiply by.
        :return: A new vector that is the product of this vector and the scalar.
        """
        return Vector3(self.x * other, self.y * other, self.z * other)

    @classmethod
    def from_list(cls, vector: List[float]) -> Self:
        """
        Factory to create a Vector3 from a list of coordinates.

        :param vector: The list of coordinates.
        :return: A new Vector3 object.
        """
        return cls(*vector)


@has_parameters
@dataclass
class Quaternion(HasParameters):
    """
    A quaternion with x, y, z and w components.
    """

    x: float = 0
    y: float = 0
    z: float = 0
    w: float = 1

    def __post_init__(self):
        self.normalize()

    def normalize(self):
        """
        Normalize the quaternion in-place.
        """
        # TODO fix this
        # if the object is not fully constructed yet
        if not (hasattr(self, "x") and
                hasattr(self, "y") and
                hasattr(self, "z") and
                hasattr(self, "w")):
            return

        norm = (self.x ** 2 + self.y ** 2 + self.z ** 2 + self.w ** 2) ** 0.5
        object.__setattr__(self, "x", self.x / norm)
        object.__setattr__(self, "y", self.y / norm)
        object.__setattr__(self, "z", self.z / norm)
        object.__setattr__(self, "w", self.w / norm)

    def ros_message(self):
        from geometry_msgs.msg import Quaternion as ROSQuaternion
        return ROSQuaternion(x=float(self.x), y=float(self.y), z=float(self.z), w=float(self.w))

    def to_list(self) -> List[float]:
        """
        Convert the quaternion to a list.

        :return: A list containing the x, y, z and w components.
        """
        return [self.x, self.y, self.z, self.w]

    def to_numpy(self) -> np.ndarray:
        """
        Convert the quaternion to a numpy array.

        :return: A numpy array containing the x, y, z and w components.
        """
        return np.array(self.to_list())

    def round(self, decimals: int = 4):
        """
        Rounds the components of the quaternion to the specified number of decimal places.

        :param decimals: The number of decimal places to round to.
        """
        self.x = round(self.x, decimals)
        self.y = round(self.y, decimals)
        self.z = round(self.z, decimals)
        self.w = round(self.w, decimals)

    def almost_equal(self, other: Self, tolerance: float = 1e-6) -> bool:
        """
        Check if two quaternions are almost equal within a given tolerance.

        :param other: The other quaternion to compare to.
        :param tolerance: The tolerance for the comparison as number of decimal places.
        :return: True if the quaternions are almost equal, False otherwise.
        """
        return bool(np.isclose(np.array(self.to_list()), np.array(other.to_list()), atol=tolerance).all())

    def __mul__(self, other: Self) -> Quaternion:
        """
        Multiplies two quaternions together.

        :param other: The other quaternion to multiply with.
        :return: A new quaternion that is the product of this quaternion and the other quaternion.
        """
        return Quaternion.from_list(quaternion_multiply(self.to_list(), other.to_list()))

    @classmethod
    def from_list(cls, quaternion: List[float]) -> Self:
        """
        Factory to create a Quaternion from a list of components.

        :param quaternion: A list of components [x, y, z, w].
        :return: A new Quaternion object.
        """
        return cls(*quaternion)

    # # TODO fix this
    # def __setattr__(self, key, value):
    #      object.__setattr__(self, key, value)
    #      self.normalize()


@has_parameters
@dataclass
class Pose(HasParameters):
    """
    A pose in 3D space.
    """
    position: Vector3 = field(default_factory=Vector3)
    orientation: Quaternion = field(default_factory=Quaternion)

    def __repr__(self):
        return (f"Pose: {[round(v, 3) for v in [self.position.x, self.position.y, self.position.z]]}, "
                f"{[round(v, 3) for v in [self.orientation.x, self.orientation.y, self.orientation.z, self.orientation.w]]}")

    def ros_message(self):
        """
        Convert the pose to a ROS message of type Pose.

        :return: The ROS message.
        """
        from geometry_msgs.msg import Point as ROSPoint
        from geometry_msgs.msg import Pose as ROSPose
        point = ROSPoint(x=float(self.position.x), y=float(self.position.y), z=float(self.position.z))
        return ROSPose(position=point, orientation=self.orientation.ros_message())

    def to_list(self):
        """
        Convert the pose to a list of [position, orientation].

        :return: A list containing the position and orientation of this pose.
        """
        return [self.position.to_list(), self.orientation.to_list()]

    def copy(self) -> Self:
        """
        Create a deep copy of the pose.

        :return: A new Pose object that is a copy of this pose.
        """
        return copy.deepcopy(self)

    def round(self, decimals: int = 4):
        """
        Rounds the components of the pose (position and orientation) to the specified number of decimal places.

        :param decimals: The number of decimal places to round to.
        """
        self.position.round(decimals)
        self.orientation.round(decimals)

    def almost_equal(self, other: Pose, position_tolerance: float = 1e-6, orientation_tolerance: float = 1e-5) -> bool:
        """
        Check if two poses are almost equal within given tolerances for position and orientation.

        :param other: The other pose to compare to.
        :param position_tolerance: Tolerance for position comparison as number of decimal places.
        :param orientation_tolerance: Tolerance for orientation comparison as number of decimal places.
        :return:  True if the poses are almost equal, False otherwise.
        """
        return self.position.almost_equal(other.position, position_tolerance) and self.orientation.almost_equal(
            other.orientation, orientation_tolerance)

    def __eq__(self, other: Self) -> bool:
        """
        Check if two poses are equal. Uses almost_equal with a tolerance of 1e-4 for both position and orientation.

        :param other: The other pose to compare to.
        :return: True if the poses are equal, False otherwise.
        """
        return self.almost_equal(other, position_tolerance=1e-4, orientation_tolerance=1e-4)

    @classmethod
    def from_matrix(cls, matrix: np.ndarray) -> Self:
        """
        Create a Pose from a 4x4 transformation matrix.

        :param matrix: A 4x4 transformation matrix as numpy array.
        :return: A pose object created from the matrix.
        """
        translation = translation_from_matrix(matrix)
        rotation = quaternion_from_matrix(matrix)
        return cls.from_list(translation, rotation)

    @classmethod
    def from_list(cls, position: List[float], orientation: List[float]) -> Self:
        """
        Factory to create a Pose from a list of position and orientation.

        :param position: List of position [x, y, z].
        :param orientation: List of orientation [x, y, z, w].
        :return: A new Pose object.
        """
        return cls(Vector3(position[0], position[1], position[2]),
                   Quaternion(orientation[0], orientation[1], orientation[2], orientation[3]))


@dataclass
class Header:
    """
    A header with a timestamp.
    """
    frame_id: str = "map"
    stamp: datetime.datetime = field(default_factory=datetime.datetime.now, compare=False)
    sequence: int = field(default=0, compare=False)

    def ros_message(self):
        """
        Convert the header to a ROS message of type Header.

        :return: The ROS message.
        """
        from std_msgs.msg import Header as ROSHeader
        split_time = str(self.stamp.timestamp()).split(".")
        stamp = ROSTime(int(split_time[0]), int(split_time[1]))
        return ROSHeader(frame_id=self.frame_id, stamp=stamp, seq=self.sequence)


@has_parameters
@dataclass
class PoseStamped(HasParameters):
    """
    A pose in 3D space with a timestamp.
    """
    pose: Pose = field(default_factory=Pose)
    header: Header = field(default_factory=Header)

    @property
    def position(self):
        return self.pose.position

    @position.setter
    def position(self, value: Vector3):
        self.pose.position = value

    @property
    def orientation(self):
        return self.pose.orientation

    @orientation.setter
    def orientation(self, value: Quaternion):
        self.pose.orientation = value

    @property
    def frame_id(self):
        return self.header.frame_id

    @frame_id.setter
    def frame_id(self, value: str):
        self.header.frame_id = value

    def __repr__(self):
        return (f"Pose: {[round(v, 3) for v in [self.position.x, self.position.y, self.position.z]]}, "
                f"{[round(v, 3) for v in [self.orientation.x, self.orientation.y, self.orientation.z, self.orientation.w]]} "
                f"in frame_id {self.frame_id}")

    def ros_message(self):
        """
        Convert the pose to a ROS message of type PoseStamped.

        :return: The ROS message.
        """
        from geometry_msgs.msg import PoseStamped as ROSPoseStamped
        return ROSPoseStamped(pose=self.pose.ros_message(), header=self.header.ros_message())

    @classmethod
    def from_ros_message(cls, message: 'ROSPoseStamped') -> Self:
        """
        Create a PoseStamped from a ROS message.

        :param message: The PoseStamped ROS message.
        :return: A new PoseStamped object created from the ROS message.
        """
        header = Header(frame_id=message.header.frame_id, stamp=message.header.stamp)
        position = Vector3(x=message.pose.position.x, y=message.pose.position.y, z=message.pose.position.z)
        orientation = Quaternion(x=message.pose.orientation.x, y=message.pose.orientation.y,
                                 z=message.pose.orientation.z, w=message.pose.orientation.w)
        return cls(pose=Pose(position=position, orientation=orientation), header=header)

    @classmethod
    def from_list(cls, position: Optional[List[float]] = None, orientation: Optional[List[float]] = None,
                  frame: Optional[str] = "map") -> Self:
        """
        Factory to create a PoseStamped from a list of position and orientation.

        :param position: Position as a list of [x, y, z].
        :param orientation: Orientation as a list of [x, y, z, w].
        :param frame: Frame in which the pose is defined.
        :return: A new PoseStamped object.
        """
        position = position or [0.0, 0.0, 0.0]
        orientation = orientation or [0.0, 0.0, 0.0, 1.0]
        return cls(pose=Pose.from_list(position, orientation),
                   header=Header(frame_id=frame, stamp=datetime.datetime.now()))

    def to_transform_stamped(self, child_link_id: str) -> TransformStamped:
        """
        Converts the PoseStamped to a TransformStamped given a frame to which the transform is pointing.

        :param child_link_id: Frame to which the transform is pointing.
        :return: A TransformStamped object.
        """
        return TransformStamped(header=self.header,
                                pose=Transform.from_list(self.position.to_list(), self.orientation.to_list()),
                                child_frame_id=child_link_id)

    def round(self, decimals: int = 4):
        """
        Rounds the components of the pose (position and orientation) to the specified number of decimal places.

        :param decimals: Number of decimal places to round to.
        """
        self.position.round(decimals)
        self.orientation.round(decimals)

    def copy(self) -> Self:
        """
        Create a deep copy of the PoseStamped object.

        :return: A new PoseStamped object that is a copy of this object.
        """
        return copy.deepcopy(self)

    def to_list(self):
        """
        Convert the pose to a list of [position, orientation, frame_id].

        :return: A list of [pose, frame_id].
        """
        return [self.pose.to_list(), self.frame_id]

    @staticmethod
    def calculate_closest_faces(pose_to_robot_vector: Vector3,
                                specified_grasp_axis: Optional[AxisIdentifier] = None) -> Tuple[
        GraspDescription, GraspDescription]:
        """
        Determines the faces of the object based on the input vector.

        If `specified_grasp_axis` is None, it calculates the primary and secondary faces based on the vector's magnitude
        determining which sides of the object are most aligned with the robot. This will either be the x, y plane for side faces
        or the z axis for top/bottom faces.
        If `specified_grasp_axis` is provided, it only considers the specified axis and calculates the faces aligned
        with that axis.

        :param pose_to_robot_vector: A 3D vector representing one of the robot's axes in the pose's frame, with
                              irrelevant components set to np.nan.
        :param specified_grasp_axis: Specifies a specific axis (e.g., X, Y, Z) to focus on.

        :return: A tuple of two Grasp enums representing the primary and secondary faces.
        """
        all_axes = [AxisIdentifier.X, AxisIdentifier.Y, AxisIdentifier.Z]

        if specified_grasp_axis:
            valid_axes = [specified_grasp_axis]
        else:
            valid_axes = [axis for axis in all_axes if
                          not np.isnan(pose_to_robot_vector.to_list()[axis.value.index(1)])]

        object_to_robot_vector = np.array(pose_to_robot_vector.to_list()) + 1e-9
        sorted_axes = sorted(valid_axes, key=lambda axis: abs(object_to_robot_vector[axis.value.index(1)]),
                             reverse=True)

        primary_axis = sorted_axes[0]
        primary_sign = int(np.sign(object_to_robot_vector[primary_axis.value.index(1)]))
        primary_face = Grasp.from_axis_direction(primary_axis, primary_sign)

        if len(sorted_axes) > 1:
            secondary_axis = sorted_axes[1]
            secondary_sign = int(np.sign(object_to_robot_vector[secondary_axis.value.index(1)]))
            secondary_face = Grasp.from_axis_direction(secondary_axis, secondary_sign)
        else:
            secondary_sign = -primary_sign
            secondary_axis = primary_axis
            secondary_face = Grasp.from_axis_direction(secondary_axis, secondary_sign)

        return primary_face, secondary_face

    def calculate_grasp_descriptions(self, robot: Object, grasp_alignment: Optional[PreferredGraspAlignment] = None) -> \
            List[GraspDescription]:
        """
        This method determines the possible grasp configurations (approach axis and vertical alignment) of the self,
        taking into account the self's orientation, position, and whether the gripper should be rotated by 90°.

        :param robot: The robot for which the grasp configurations are being calculated.

        :return: A sorted list of GraspDescription instances representing all grasp permutations.
        """
        objectTmap = self

        robot_pose = robot.get_pose()

        if grasp_alignment:
            side_axis = grasp_alignment.preferred_axis
            vertical = grasp_alignment.with_vertical_alignment
            rotated_gripper = grasp_alignment.with_rotated_gripper
        else:
            side_axis, vertical, rotated_gripper = None, False, False

        object_to_robot_vector_world = objectTmap.position.vector_to_position(robot_pose.position)
        orientation = objectTmap.orientation.to_list()

        mapRobject = R.from_quat(orientation).as_matrix()
        objectRmap = mapRobject.T

        object_to_robot_vector_local = objectRmap.dot(object_to_robot_vector_world.to_numpy())
        vector_x, vector_y, vector_z = object_to_robot_vector_local

        vector_side = Vector3(vector_x, vector_y, np.nan)
        side_faces = self.calculate_closest_faces(vector_side, side_axis)

        vector_vertical = Vector3(np.nan, np.nan, vector_z)
        vertical_faces = self.calculate_closest_faces(vector_vertical) if vertical else [None]

        grasp_configs = [
            GraspDescription(approach_direction=side, vertical_alignment=top_face, rotate_gripper=rotated_gripper)
            for top_face in vertical_faces
            for side in side_faces
        ]

        return grasp_configs

    def almost_equal(self, other: PoseStamped, position_tolerance: float = 1e-6,
                     orientation_tolerance: float = 1e-5) -> bool:
        """
        Check if two PoseStamped objects are almost equal within given tolerances for position and orientation and if the
        frame_id is the same.

        :param other: The other PoseStamped object to compare to.
        :param position_tolerance: Tolerance for position comparison as number of decimal places.
        :param orientation_tolerance: Tolerance for orientation comparison as number of decimal places.
        :return: True if the PoseStamped objects are almost equal, False otherwise.
        """
        return self.position.almost_equal(other.position, position_tolerance) and self.orientation.almost_equal(
            other.orientation, orientation_tolerance) and self.frame_id == other.frame_id

    def rotate_by_quaternion(self, quaternion: List[float]):
        """
        Rotates the orientation of the pose by a given quaternion.

        :param quaternion: A list representing the quaternion [x, y, z, w].
        """
        self.orientation = self.orientation * Quaternion.from_list(quaternion)

    def is_facing_2d_axis(self, pose_b: PoseStamped, axis: Optional[AxisIdentifier] = AxisIdentifier.X,
                          threshold_deg=82) -> \
            Tuple[bool, float]:
        """
         Check if this pose is facing another pose along a specific axis (X or Y) within a given angular threshold.

         :param pose_b: The target pose to compare against.
         :param axis: The axis to check alignment with ('x' or 'y'). Defaults to 'x'.
         :param threshold_deg: The maximum angular difference in degrees to consider as 'facing'. Defaults to 82 degrees.
         :return: Tuple of (True/False if facing, signed angular difference in radians).
         """

        a_pos = np.array([self.position.x, self.position.y])
        b_pos = np.array([pose_b.position.x, pose_b.position.y])
        target_angle = math.atan2(*(b_pos - a_pos)[::-1])

        q = self.orientation
        yaw = math.atan2(2 * (q.w * q.z + q.x * q.y), 1 - 2 * (q.y ** 2 + q.z ** 2))
        if axis == AxisIdentifier.Y:
            yaw += math.pi / 2
        diff = math.atan2(math.sin(target_angle - yaw), math.cos(target_angle - yaw))
        return abs(diff) < math.radians(threshold_deg), diff

    def is_facing_x_or_y(self, pose_b: PoseStamped) -> bool:
        """
        Check if this pose is facing another pose along either the X or Y axis within a default angular threshold.

        :param pose_b: The target pose to compare against.
        :return: True if this pose is facing the target along either X or Y axis, False otherwise.
        """

        result = {a: self.is_facing_2d_axis(pose_b, axis=a) for a in (AxisIdentifier.X, AxisIdentifier.Y)}
        return any(r[0] for r in result.values())



@dataclass
class Transform(Pose):
    @property
    def translation(self):
        return self.position

    @property
    def rotation(self):
        return self.orientation

    def to_matrix(self) -> np.ndarray:
        """
        Converts the transform to a 4x4 transformation matrix.

        :return: A numpy array representing the transformation matrix.
        """
        translation = translation_matrix(self.translation.to_list())
        rotation = quaternion_matrix(self.rotation.to_list())
        return np.dot(translation, rotation)

    def __invert__(self) -> Transform:
        """
        Inverts the transform, returning a new Transform object.

        :return: The inverted Transform object.
        """
        inv = inverse_matrix(self.to_matrix())
        translation = translation_from_matrix(inv)
        rotation = quaternion_from_matrix(inv)
        return Transform.from_list(translation, rotation)

    def __mul__(self, other: Transform) -> Transform:
        """
        Multiplies two transforms together, returning a new Transform object.

        :param other: The other Transform object to multiply with.
        :return: A new Transform object that is the product of this transform and the other transform.
        """
        self_matrix = self.to_matrix()
        other_matrix = other.to_matrix()
        multiplication = self_matrix @ other_matrix
        return Transform.from_matrix(multiplication)

    @classmethod
    def from_pose(cls, pose: Pose) -> Self:
        """
        Create a Transform from a Pose object.

        :param pose: The pose to convert to a Transform.
        :return: A new Transform object created from the Pose.
        """
        return cls(pose.position, pose.orientation)

    def ros_message(self):
        """
        Convert the transform to a ROS message of type Transform.

        :return: The ROS message.
        """
        from geometry_msgs.msg import Transform as ROSTransform
        return ROSTransform(translation=self.translation.ros_message(), rotation=self.rotation.ros_message())


@has_parameters
@dataclass
class TransformStamped(PoseStamped):
    child_frame_id: str = field(default_factory=str)
    """
    Target frame id of the transform.
    """
    pose: Transform = field(default_factory=Pose)
    """
    The transform of the transform.
    """

    @property
    def transform(self) -> Transform:
        return self.pose

    @transform.setter
    def transform(self, value: Transform):
        self.pose = value

    @property
    def translation(self):
        return self.pose.position

    @translation.setter
    def translation(self, value: Vector3):
        self.pose.position = value

    @property
    def rotation(self):
        return self.pose.orientation

    @rotation.setter
    def rotation(self, value: Quaternion):
        self.pose.orientation = value

    def __invert__(self) -> Self:
        """
        Inverts the transform, returning a new TransformStamped object which points from child_frame_id to frame_id.

        :return: A new TransformStamped object that is the inverse of this transform.
        """
        result = copy.deepcopy(self)
        result.header.frame_id = self.child_frame_id
        result.child_frame_id = self.header.frame_id
        result.transform = ~self.transform
        return result

    def __mul__(self, other) -> Self:
        """
        Multiplies two TransformStamped objects together, returning a new TransformStamped object.

        :param other: The other TransformStamped object to multiply with.
        :return: A new TransformStamped object that is the product of this transform and the other transform.
        """
        result = copy.deepcopy(self)
        result.child_frame_id = other.child_frame_id
        result.transform = self.transform * other.transform
        return result

    @classmethod
    def from_list(cls, translation: List[float] = None, rotation: List[float] = None, frame: str = "map",
                  child_frame_id="") -> Self:
        """
        Factory to create a TransformStamped from a list of position and orientation.

        :param translation: Translation as a list of [x, y, z].
        :param rotation: Rotation as a list of [x, y, z, w].
        :param frame: Original frame in which the transform is defined.
        :param child_frame_id: Target frame id of the transform.
        :return: A new TransformStamped object.
        """
        translation = translation or [0.0, 0.0, 0.0]
        rotation = rotation or [0.0, 0.0, 0.0, 1.0]
        return cls(pose=Transform.from_list(translation, rotation),
                   header=Header(frame_id=frame, stamp=datetime.datetime.now()),
                   child_frame_id=child_frame_id)

    def ros_message(self):
        """
        Convert the TransformStamped to a ROS message of type TransformStamped.

        :return: The ROS message.
        """
        from geometry_msgs.msg import TransformStamped as ROSTransformStamped
        return ROSTransformStamped(transform=self.transform.ros_message(), header=self.header.ros_message(),
                                   child_frame_id=self.child_frame_id)

    def to_pose_stamped(self) -> PoseStamped:
        """
        Converts the TransformStamped to a PoseStamped object.

        :return: A PoseStamped object created from the TransformStamped.
        """
        p = Pose(self.pose.position, self.pose.orientation)
        return PoseStamped(p, self.header)

    def inverse_times(self, other: TransformStamped) -> Self:
        """
        Multiply this TransformStamped by the inverse of another TransformStamped. Essentially, this is the same as
        "subtracting" another TransformStamped from this one.

        :param other: The other TransformStamped to subtract.
        :return: A new TransformStamped object that is the result of this transform minus the other transform.
        """
        return self * ~other


@has_parameters
@dataclass
class GraspPose(PoseStamped):
    """
    A pose from which a grasp can be performed along with the respective arm and grasp description.
    """
    arm: Arms = None
    """
    Arm corresponding to the grasp pose.
    """
    grasp_description: GraspDescription = None
    """
    Grasp description corresponding to the grasp pose.
    """


Point = Vector3



from threading import Lock

from .default_process_modules import *
from ..designators.motion_designator import *


class KevinMoveArmJoints(DefaultMoveArmJoints):
    """
    Process module for the simulated Kevin that moves the arm joints of the robot
    """

    def _execute(self, desig: MoveArmJointsMotion):
        # Kevin has very long fingers. It is hard to grasp things with we allow collisions for the gripper when moving
        # for now
        giskard.allow_gripper_collision("gripper")
        DefaultMoveArmJoints._execute(desig)

###########################################################
########## Process Modules for the Real Kevin ###########
###########################################################

# We never had the real robot or worked on it


class KevinManager(DefaultManager):
    def __init__(self):
        super().__init__()
        self.robot_name = "kevin"
        self._navigate_lock = Lock()

    def navigate(self):
        if ProcessModuleManager.execution_type == ExecutionType.SIMULATED:
            return DefaultNavigation(self._navigate_lock)
        elif ProcessModuleManager.execution_type == "real":
            raise NotImplemented
            # return KevinNavigationReal(self._navigate_lock)

    def detecting(self):
        if ProcessModuleManager.execution_type == ExecutionType.SIMULATED:
            return DefaultDetecting(self._detecting_lock)
        elif ProcessModuleManager.execution_type == "real":
            raise NotImplemented
            # return KevinDetectingReal(self._detecting_lock)

    def move_tcp(self):
        if ProcessModuleManager.execution_type == ExecutionType.SIMULATED:
            return DefaultMoveTCP(self._move_tcp_lock)
        elif ProcessModuleManager.execution_type == "real":
            raise NotImplemented
            # return KevinMoveTCPReal(self._move_tcp_lock)

    def move_arm_joints(self):
        if ProcessModuleManager.execution_type == ExecutionType.SIMULATED:
            return DefaultMoveArmJoints(self._move_arm_joints_lock)
        elif ProcessModuleManager.execution_type == ExecutionType.REAL:
            return DefaultMoveArmJointsReal(self._move_arm_joints_lock)
            # return KevinMoveArmJointsReal(self._move_arm_joints_lock)

    def world_state_detecting(self):
        if ProcessModuleManager.execution_type == ExecutionType.SIMULATED or ProcessModuleManager.execution_type == ExecutionType.REAL:
            return DefaultWorldStateDetecting(self._world_state_detecting_lock)

    def move_joints(self):
        if ProcessModuleManager.execution_type == ExecutionType.SIMULATED:
            return DefaultMoveJoints(self._move_arm_joints_lock)
        elif ProcessModuleManager.execution_type == ExecutionType.REAL:
            return DefaultMoveJointsReal(self._move_arm_joints_lock)

    def move_gripper(self):
        if ProcessModuleManager.execution_type == ExecutionType.SIMULATED:
            return DefaultMoveGripper(self._move_gripper_lock)
        elif ProcessModuleManager.execution_type == "real":
            raise NotImplemented
            # return KevinMoveGripperReal(self._move_gripper_lock)

KevinManager()


"""Implementation of process modules.

Classes:
ProcessModule -- implementation of process modules.
"""
# used for delayed evaluation of typing until python 3.11 becomes mainstream
from __future__ import annotations

import glob
import importlib
import inspect
from os.path import dirname, basename, isfile, join
from dataclasses import field
from datetime import timedelta
from threading import Lock, get_ident
import time
from abc import ABC
from typing_extensions import Callable, Any, Union, Optional, List

from .robot_description import RobotDescription
from .datastructures.world import World
from typing_extensions import TYPE_CHECKING
from .datastructures.enums import ExecutionType
from .ros import logerr, logwarn_once

if TYPE_CHECKING:
    from .designators.motion_designator import BaseMotion


class ProcessModule:
    """
    Implementation of process modules. Process modules are the part that communicate with the outer world to execute
     designators.
    """
    execution_delay: Optional[timedelta] = World.conf.execution_delay
    """
    Adds a delay after executing a process module, to make the execution in simulation more realistic
    """

    def __init__(self, lock):
        """Create a new process module."""
        self._lock = lock

    def _execute(self, designator: BaseMotion) -> Any:
        """
        Helper method for internal usage only.
        This method is to be overwritten instead of the execute method.
        """
        pass

    def execute(self, designator: BaseMotion) -> Any:
        """
        Execute the given designator_description. If there is already another process module of the same kind the `self._lock` will
        lock this thread until the execution of that process module is finished. This implicitly queues the execution of
        process modules.

        :param designator: The designator_description to execute.
        :return: Return of the Process Module if there is any
        """
        with self._lock:
            ret = self._execute(designator)
            if ProcessModule.execution_delay:
                time.sleep(self.execution_delay.total_seconds())

        return ret


class RealRobot:
    """
    Management class for executing designators on the real robot. This is intended to be used in a with environment.
    When importing this class an instance is imported instead.

    Example:

    .. code-block:: python

        with real_robot:
            some designators
    """

    def __init__(self):
        self.pre: ExecutionType = ExecutionType.REAL
        self.pre_delay: timedelta = World.conf.execution_delay

    def __enter__(self):
        """
        Entering function for 'with' scope, saves the previously set :py:attr:`~ProcessModuleManager.execution_type` and
        sets it to 'real'
        """
        self.pre = ProcessModuleManager.execution_type
        ProcessModuleManager.execution_type = ExecutionType.REAL
        self.pre_delay = ProcessModule.execution_delay
        ProcessModule.execution_delay = timedelta(seconds=0)

    def __exit__(self, _type, value, traceback):
        """
        Exit method for the 'with' scope, sets the :py:attr:`~ProcessModuleManager.execution_type` to the previously
        used one.
        """
        ProcessModuleManager.execution_type = self.pre
        ProcessModule.execution_delay = self.pre_delay

    def __call__(self):
        return self


class SimulatedRobot:
    """
    Management class for executing designators on the simulated robot. This is intended to be used in
    a with environment. When importing this class an instance is imported instead.

    Example:

    .. code-block:: python

        with simulated_robot:
            some designators
    """

    def __init__(self):
        self.pre: ExecutionType = ExecutionType.SIMULATED

    def __enter__(self):
        """
        Entering function for 'with' scope, saves the previously set :py:attr:`~ProcessModuleManager.execution_type` and
        sets it to 'simulated'
        """
        self.pre = ProcessModuleManager.execution_type
        ProcessModuleManager.execution_type = ExecutionType.SIMULATED

    def __exit__(self, _type, value, traceback):
        """
        Exit method for the 'with' scope, sets the :py:attr:`~ProcessModuleManager.execution_type` to the previously
        used one.
        """
        ProcessModuleManager.execution_type = self.pre

    def __call__(self):
        return self


class SemiRealRobot:
    """
    Management class for executing designators on the semi-real robot. This is intended to be used in a with environment.
    When importing this class an instance is imported instead.

    Example:

    .. code-block:: python

        with semi_real_robot:
            some designators
    """

    def __init__(self):
        self.pre: ExecutionType = ExecutionType.SEMI_REAL

    def __enter__(self):
        """
        Entering function for 'with' scope, saves the previously set :py:attr:`~ProcessModuleManager.execution_type` and
        sets it to 'semi_real'
        """
        self.pre = ProcessModuleManager.execution_type
        ProcessModuleManager.execution_type = ExecutionType.SEMI_REAL

    def __exit__(self, type, value, traceback):
        """
        Exit method for the 'with' scope, sets the :py:attr:`~ProcessModuleManager.execution_type` to the previously
        used one.
        """
        ProcessModuleManager.execution_type = self.pre

    def __call__(self):
        return self


def with_real_robot(func: Callable) -> Callable:
    """
    Decorator to execute designators in the decorated class on the real robot.

    Example:

    .. code-block:: python

        @with_real_robot
        def plan():
            some designators

    :param func: Function this decorator is annotating
    :return: The decorated function wrapped into the decorator
    """

    def wrapper(*args, **kwargs):
        pre = ProcessModuleManager.execution_type
        ProcessModuleManager.execution_type = ExecutionType.REAL
        ret = func(*args, **kwargs)
        ProcessModuleManager.execution_type = pre
        return ret

    return wrapper


def with_simulated_robot(func: Callable) -> Callable:
    """
    Decorator to execute designators in the decorated class on the simulated robot.

    Example:

    .. code-block:: python

        @with_simulated_robot
        def plan():
            some designators

    :param func: Function this decorator is annotating
    :return: The decorated function wrapped into the decorator
    """

    def wrapper(*args, **kwargs):
        pre = ProcessModuleManager.execution_type
        ProcessModuleManager.execution_type = ExecutionType.SIMULATED
        ret = func(*args, **kwargs)
        ProcessModuleManager.execution_type = pre
        return ret

    return wrapper


# These are imported, so they don't have to be initialized when executing with
simulated_robot = SimulatedRobot()
real_robot = RealRobot()
semi_real_robot = SemiRealRobot()


class ProcessModuleManager(ABC):
    """
    Base class for managing process modules, any new process modules have to implement this class to register the
    Process Modules
    """
    execution_type: ExecutionType = None
    """
    Whether the robot for which the process module is intended for is real or a simulated one
    """
    available_pms: List[ProcessModuleManager] = []
    """
    List of all available Process Module Managers
    """
    _instance: ProcessModuleManager = None
    """
    Singelton instance of this Process Module Manager
    """

    def __new__(cls, *args, **kwargs):
        """
        Creates a new instance if :py:attr:`~ProcessModuleManager._instance` is None, otherwise the instance
        in :py:attr:`~ProcessModuleManager._instance` is returned.

        :return: Singelton instance of this Process Module Manager
        """
        if not cls._instance:
            cls._instance = super(ProcessModuleManager, cls).__new__(cls)
            return cls._instance
        else:
            return cls._instance

    def __init__(self, robot_name: str):
        """
        Registers the Process modules for this robot. The name of the robot has to match the name given in the robot
        description.

        :param robot_name: Name of the robot for which these Process Modules are intended
        """

        self.robot_name = robot_name
        self.available_pms.append(self)
        self._navigate_lock = Lock()
        self._pick_up_lock = Lock()
        self._place_lock = Lock()
        self._looking_lock = Lock()
        self._detecting_lock = Lock()
        self._move_tcp_lock = Lock()
        self._move_arm_joints_lock = Lock()
        self._world_state_detecting_lock = Lock()
        self._move_joints_lock = Lock()
        self._move_gripper_lock = Lock()
        self._open_lock = Lock()
        self._close_lock = Lock()
        self._move_tcp_waypoints_lock = Lock()
        self.available_pms = []
        self.register_all_process_modules()

    @staticmethod
    def get_manager() -> Union[ProcessModuleManager, None]:
        """
        Returns the Process Module manager for the currently loaded robot or None if there is no Manager.

        :return: ProcessModuleManager instance of the current robot
        """
        ProcessModuleManager.register_all_process_modules()
        manager = None
        _default_manager = None
        if not ProcessModuleManager.execution_type:
            raise RuntimeError(
                f"No execution_type is set, did you use the with_simulated_robot or with_real_robot decorator?")

        robot_description = RobotDescription.current_robot_description
        chains = robot_description.get_manipulator_chains()
        gripper_name = [chain.end_effector.gripper_object_name for chain in chains
                        if chain.end_effector.gripper_object_name]
        gripper_name = gripper_name[0] if len(gripper_name) > 0 else None

        for pm_manager in ProcessModuleManager.available_pms:
            if pm_manager.robot_name == robot_description.name or\
                    ((pm_manager.robot_name == gripper_name) and gripper_name):
                manager = pm_manager
            if pm_manager.robot_name == "default":
                _default_manager = pm_manager

        if manager:
            return manager
        elif _default_manager:
            logwarn_once(f"No Process Module Manager found for robot: '{RobotDescription.current_robot_description.name}'"
                               f", using default process modules")
            return _default_manager
        else:
            logerr(f"No Process Module Manager found for robot: '{RobotDescription.current_robot_description.name}'"
                         f", and no default process modules available")
            return None

    @staticmethod
    def register_all_process_modules():
        modules = glob.glob(join(dirname(__file__) + "/process_modules", "*.py"))
        __all__ = [basename(f)[:-3] for f in modules if isfile(f) and not f.endswith('__init__.py')]

        for module_name in __all__:
            try:
                importlib.import_module(f".{module_name}", package="pycram.process_modules")
            except Exception as e:
                print(f"Error loading module {module_name}: {e}")

    def navigate(self) -> ProcessModule:
        """
        Get the Process Module for navigating the robot with respect to
         the :py:attr:`~ProcessModuleManager.execution_type`

        :return: The Process Module for navigating
        """
        raise NotImplementedError(
            f"There are no Process Modules for '{inspect.currentframe().f_code.co_name}' for robot '{self.robot_name}'")

    def pick_up(self) -> ProcessModule:
        """
        Get the Process Module for picking up with respect to the :py:attr:`~ProcessModuleManager.execution_type`

        :return: The Process Module for picking up an object
        """
        raise NotImplementedError(
            f"There are no Process Modules for '{inspect.currentframe().f_code.co_name}' for robot '{self.robot_name}'")

    def place(self) -> ProcessModule:
        """
        Get the Process Module for placing with respect to the :py:attr:`~ProcessModuleManager.execution_type`

        :return: The Process Module for placing an Object
        """
        raise NotImplementedError(
            f"There are no Process Modules for '{inspect.currentframe().f_code.co_name}' for robot '{self.robot_name}'")

    def looking(self) -> ProcessModule:
        """
        Get the Process Module for looking at a point with respect to
         the :py:attr:`~ProcessModuleManager.execution_type`

        :return: The Process Module for looking at a specific point
        """
        raise NotImplementedError(
            f"There are no Process Modules for '{inspect.currentframe().f_code.co_name}' for robot '{self.robot_name}'")

    def detecting(self) -> ProcessModule:
        """
        Get the Process Module for detecting an object with respect to
         the :py:attr:`~ProcessModuleManager.execution_type`

        :return: The Process Module for detecting an object
        """
        raise NotImplementedError(
            f"There are no Process Modules for '{inspect.currentframe().f_code.co_name}' for robot '{self.robot_name}'")

    def move_tcp(self) -> ProcessModule:
        """
        Get the Process Module for moving the Tool Center Point with respect to
         the :py:attr:`~ProcessModuleManager.execution_type`

        :return: The Process Module for moving the TCP
        """
        raise NotImplementedError(
            f"There are no Process Modules for '{inspect.currentframe().f_code.co_name}' for robot '{self.robot_name}'")

    def move_arm_joints(self) -> ProcessModule:
        """
        Get the Process Module for moving the joints of the robot arm
        with respect to the :py:attr:`~ProcessModuleManager.execution_type`

        :return: The Process Module for moving the arm joints
        """
        raise NotImplementedError(
            f"There are no Process Modules for '{inspect.currentframe().f_code.co_name}' for robot '{self.robot_name}'")

    def world_state_detecting(self) -> ProcessModule:
        """
        Get the Process Module for detecting an object using the world state with respect to the
        :py:attr:`~ProcessModuleManager.execution_type`

        :return: The Process Module for world state detecting
        """
        raise NotImplementedError(
            f"There are no Process Modules for '{inspect.currentframe().f_code.co_name}' for robot '{self.robot_name}'")

    def move_joints(self) -> ProcessModule:
        """
        Get the Process Module for moving any joint of the robot with respect to the
        :py:attr:`~ProcessModuleManager.execution_type`

        :return: The Process Module for moving joints
        """
        raise NotImplementedError(
            f"There are no Process Modules for '{inspect.currentframe().f_code.co_name}' for robot '{self.robot_name}'")

    def move_gripper(self) -> ProcessModule:
        """
        Get the Process Module for moving the gripper with respect to
         the :py:attr:`~ProcessModuleManager.execution_type`

        :return: The Process Module for moving the gripper
        """
        raise NotImplementedError(
            f"There are no Process Modules for '{inspect.currentframe().f_code.co_name}' for robot '{self.robot_name}'")

    def open(self) -> ProcessModule:
        """
        Get the Process Module for opening drawers with respect to
         the :py:attr:`~ProcessModuleManager.execution_type`

        :return: The Process Module for opening drawers
        """
        raise NotImplementedError(
            f"There are no Process Modules for '{inspect.currentframe().f_code.co_name}' for robot '{self.robot_name}'")

    def close(self) -> ProcessModule:
        """
        Get the Process Module for closing drawers with respect to
         the :py:attr:`~ProcessModuleManager.execution_type`

        :return: The Process Module for closing drawers
        """
        raise NotImplementedError(
            f"There are no Process Modules for '{inspect.currentframe().f_code.co_name}' for robot '{self.robot_name}'")

    def move_tcp_waypoints(self) -> ProcessModule:
        """
        Get the Process Module for moving the Tool Center Point along a list of waypoints with respect to
         the :py:attr:`~ProcessModuleManager.execution_type`

        :return: The Process Module for moving the TCP along a list of waypoints
        """
        raise NotImplementedError(
            f"There are no Process Modules for '{inspect.currentframe().f_code.co_name}' for robot '{self.robot_name}'")


"""Implementation of helper functions and classes for internal usage only.

Classes:
Singleton -- implementation of singleton metaclass
"""
import os
from typing_extensions import Dict, Optional, Tuple
import xml.etree.ElementTree as ET

from .datastructures.enums import DescriptionType
from .object_descriptors.urdf import ObjectDescription as URDFObject
from .ros import  loginfo


class Singleton(type):
    """
    Metaclass for singletons
    """

    _instances = {}
    """
    Dictionary of singleton child classes inheriting from this metaclass, keyed by child class objects.
    """

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


def get_robot_urdf_and_mjcf_file_paths(robot_name: str, robot_relative_dir: str,
                                       multiverse_resources: Optional[str] = None)\
        -> Tuple[Optional[str], Optional[str]]:
    """
    Get the paths to the MJCF and URDF files of a robot from the Multiverse resources directory.

    :param robot_name: The name of the robot.
    :param robot_relative_dir: The relative directory of the robot in the Multiverse resources/robots directory.
    :param multiverse_resources: The path to the Multiverse resources directory.
    """
    multiverse_resources = find_multiverse_resources_path() if multiverse_resources is None else multiverse_resources

    urdf_filename: Optional[str] = None
    mjcf_filename: Optional[str] = None
    if multiverse_resources is not None:
        urdf_filename = get_robot_description_path(robot_relative_dir, robot_name,
                                                   description_type=URDFObject,
                                                   resources_dir=multiverse_resources)
        mjcf_filename = get_robot_description_path(robot_relative_dir, robot_name,
                                                   resources_dir=multiverse_resources)
    return urdf_filename, mjcf_filename


def parse_mjcf_actuators(file_path: str) -> Dict[str, str]:
    """
    Parse the actuator elements from an MJCF file.

    :param file_path: The path to the MJCF file.
    """
    tree = ET.parse(file_path)
    root = tree.getroot()

    joint_actuators = {}

    # Iterate through all actuator elements
    for actuator in root.findall(".//actuator/*"):
        name = actuator.get('name')
        joint = actuator.get('joint')
        if name and joint:
            joint_actuators[joint] = name

    return joint_actuators


def get_robot_description_path(robot_relative_dir: str, robot_name: str,
                               description_type: DescriptionType = DescriptionType.MJCF,
                               file_name: Optional[str] = None,
                               resources_dir: Optional[str] = None) -> Optional[str]:
    """
    Get the path to the description file of a robot.

    :param robot_relative_dir: The relative directory of the robot in the resources/robots directory.
    :param robot_name: The name of the robot.
    :param description_type: The type of the description (URDF, MJCF).
    :param file_name: The name of the XML file of the robot.
    :param resources_dir: The path to the Multiverse resources directory.
    :return: The path to the description file of the robot if it exists, otherwise None.
    """
    resources_dir = find_multiverse_resources_path() if resources_dir is None else resources_dir
    if resources_dir is None:
        loginfo("Could not find Multiverse resources path and no other resources were given.")
        return None
    file_name = file_name if file_name is not None else robot_name
    extension = description_type.get_file_extension()
    extension_folders = []
    if extension[1:] in ['xml', 'mjcf']:
        extension_folders.append('mjcf')
    else:
        extension_folders.append(extension[1:])
    if extension not in file_name:
        file_name = file_name + extension

    robot_folders = [os.path.join(resources_dir, 'robots', robot_relative_dir)]
    robot_folders.append(os.path.join(robot_folders[0], robot_name))
    for robot_folder in robot_folders:
        if resources_dir is not None and os.path.exists(robot_folder):
            list_dir = os.listdir(robot_folder)
            for extension_folder in extension_folders:
                if extension_folder in list_dir:
                    if file_name in os.listdir(robot_folder + f'/{extension_folder}'):
                        return os.path.join(robot_folder, extension_folder, file_name)
                elif file_name in os.listdir(robot_folder):
                    return os.path.join(robot_folder, file_name)
    logwarn(f"Robot {robot_name} not found in resources.")
    return None


def find_multiverse_resources_path() -> Optional[str]:
    """
    :return: The path to the Multiverse resources directory.
    """
    # Get the path to the Multiverse installation
    multiverse_path = find_multiverse_path()

    # Check if the path to the Multiverse installation was found
    if multiverse_path:
        # Construct the path to the resources directory
        resources_path = os.path.join(multiverse_path, 'resources')

        # Check if the resources directory exists
        if os.path.exists(resources_path):
            return resources_path

    return None


def find_multiverse_path() -> Optional[str]:
    """
    :return: the path to the Multiverse installation.
    """
    # Get the value of PYTHONPATH environment variable
    pythonpath = os.getenv('PYTHONPATH')
    multiverse_relative_path = "Multiverse/multiverse"

    # Check if PYTHONPATH is set
    if pythonpath:
        # Split the PYTHONPATH into individual paths using the platform-specific path separator
        paths = pythonpath.split(os.pathsep)

        # Iterate through each path and check if 'Multiverse' is in it
        for path in paths:
            if multiverse_relative_path in path:
                multiverse_path = path.split(multiverse_relative_path)[0]
                return multiverse_path + multiverse_relative_path


def perform(action_instance):
    """
    Executes the perform logic for a given action instance.

    :param action_instance: An instance of an action class.
    """
    return action_instance.perform()


def an(designator):
    """
    Resolve the first available action from the designator.

    :param designator: The designator description instance.
    :return: The first resolved action instance.
    """
    return designator.resolve()



import time
import threading
import atexit

from ..datastructures.pose import PoseStamped
from ..datastructures.world import World
from ..datastructures.enums import ExecutionType
from tf2_msgs.msg import TFMessage

from ..ros import  create_publisher
from ..ros import  Time


class TFBroadcaster:
    """
    Broadcaster that publishes TF frames for every object in the World.
    """
    def __init__(self, projection_namespace=ExecutionType.SIMULATED, odom_frame="odom", interval=0.1):
        """
        The broadcaster prefixes all published TF messages with a projection namespace to distinguish between the TF
        frames from the simulation and the one from the real robot.

        :param projection_namespace: Name with which the TF frames should be prefixed
        :param odom_frame: Name of the statically published odom frame
        :param interval: Interval at which the TFs should be published, in seconds
        """
        self.world = World.current_world

        self.tf_static_publisher = create_publisher("/tf_static", TFMessage, queue_size=10)
        self.tf_publisher = create_publisher("/tf", TFMessage, queue_size=10)
        self.thread = threading.Thread(target=self._publish, daemon=True)
        self.kill_event = threading.Event()
        self.interval = interval

        # Namespaces
        self.projection_namespace = projection_namespace
        self.odom_frame = odom_frame

        self.thread.start()

        atexit.register(self._stop_publishing)

    def update(self):
        """
        Updates the TFs for the static odom frame and all objects currently in the World.
        """
        # Update static odom
        self._update_static_odom()
        # Update pose of objects which are possibly attached on the robot
        self._update_objects()

    def _update_objects(self) -> None:
        """
        Publishes the current pose of all objects in the World. As well as the poses of all links of these objects.
        """
        for obj in self.world.objects:
            pose = obj.get_pose()
            pose.header.stamp = Time().now()
            self._publish_pose(obj.tf_frame, pose)
            for link in obj.link_name_to_id.keys():
                link_pose = obj.get_link_pose(link)
                link_pose.header.stamp = Time().now()
                self._publish_pose(obj.get_link_tf_frame(link), link_pose)

    def _update_static_odom(self) -> None:
        """
        Publishes a static odom frame to the tf_static topic.
        """
        self._publish_pose(self.odom_frame,
                           PoseStamped.from_list([0, 0, 0], [0, 0, 0, 1]), static=True)

    def _publish_pose(self, child_frame_id: str, pose: PoseStamped, static=False) -> None:
        """
        Publishes the given pose to the ROS TF topic. First the pose is converted to a Transform between pose.frame and
        the given child_frame_id. Afterward, the frames of the Transform are prefixed with the projection namespace.

        :param child_frame_id: Name of the TF frame which the pose points to
        :param pose: Pose that should be published
        :param static: If the pose should be published to the tf_static topic
        """
        frame_id = pose.frame_id
        if frame_id != child_frame_id:
            tf_stamped = pose.to_transform_stamped(child_frame_id)
            tf_stamped.frame_id = self.projection_namespace.name + "/" + tf_stamped.frame_id
            tf_stamped.child_frame_id = self.projection_namespace.name + "/" + tf_stamped.child_frame_id
            tf2_msg = TFMessage()
            tf2_msg.transforms.append(tf_stamped)
            if static:
                self.tf_static_publisher.publish(tf2_msg)
            else:
                self.tf_publisher.publish(tf2_msg)

    def _publish(self) -> None:
        """
        Constantly publishes the positions of all objects in the World.
        """
        while not self.kill_event.is_set():
            self.update()
            time.sleep(self.interval)

    def _stop_publishing(self) -> None:
        """
        Called when the process ends, sets the kill_event which terminates the thread that publishes to the TF topic.
        """
        self.kill_event.set()
        self.thread.join()
